import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.formula.api as smf
import warnings

from .regression import plot_partreg
from .regression import plot_pred_bands
from .regression import plot_projreg
from .regression import plot_reg



def fit_and_plot_reg(xs, ys, ax=None):
    """
    Fit a linear model `y ~ 1 + x` for the data provided in `xs` and `ys`,
    then plot the best-fitting line on top of a scatter plot `ys` vs. `xs`.
    Returns the linear regression result object and the plot axis.
    """
    xname = xs.name if hasattr(xs, "name") and xs.name is not None else "x"
    yname = ys.name if hasattr(ys, "name") and ys.name is not None else "y"
    xydf = pd.DataFrame({xname: xs, yname: ys})
    lmres = smf.ols(f"{yname} ~ 1 + {xname}", data=xydf).fit()
    ax = plot_reg(lmres, ax=ax)
    return lmres, ax



# Old linear model plotting functions (to be deprecated in next release)
################################################################################

def plot_lm_simple(xs, ys, ax=None, ci_mean=False, alpha_mean=0.1, lab_mean=True,
                   ci_obs=False, alpha_obs=0.1, lab_obs=True): 
    """
    Draw a scatter plot of the data `[xs,ys]`, a regression line,
    and optionally show confidence intervals for the model predcitions.
    If `ci_mean` is True: draw a (1-alpha_mean)-CI for the mean.
    If `ci_obs` is True: draw a (1-ci_obs)-CI for the predicted values.
    """
    warnings.warn("This function is replaced by plot_reg and plot_pred_bands", DeprecationWarning)
    ax = plt.gca() if ax is None else ax

    # Prepare the data
    xname = xs.name if hasattr(xs, "name") else "x"
    yname = ys.name if hasattr(ys, "name") else "y"
    data = pd.DataFrame({xname: xs, yname: ys})

    # Fit the linear model
    formula = f"{yname} ~ 1 + {xname}"
    lm = smf.ols(formula, data=data).fit()

    plot_reg(lm, ax=ax)
    plot_pred_bands(lm, ax=ax,
                    ci_mean=ci_mean, alpha_mean=alpha_mean, lab_mean=lab_mean,
                    ci_obs=ci_obs, alpha_obs=alpha_obs, lab_obs=lab_obs)
    return ax


def plot_lm_partial_old(lmfit, pred, others=None, ax=None):
    """
    Generate a partial regression plot from the best-fit line
    of the predictor `pred`, where the intercept is calculated
    from the average of the `other` predictors.
    """
    warnings.warn("This function is replaced by plot_projreg", DeprecationWarning)
    return plot_projreg(lmfit, pred, others=others, ax=ax)


def plot_lm_partial(lmfit, pred, others=None, ax=None):
    """
    Generate a partial regression plot from the model `lmfit`
    for the predictor `pred`, given the `other` predictors.
    We plot the residuals of `outcome ~ other` along the y-axis,
    and the residuals of the model `pred ~ other` on the x-axis.
    """
    warnings.warn("This function is replaced by plot_partreg", DeprecationWarning)
    return plot_partreg(lmfit, pred=pred, ax=ax)



# DEPRECATED SINCE `plot_partreg` subtracts all other vars
def plot_projreg_cat(lmfit, pred, others=None, color="C0", linestyle="solid", cats=None, ax=None):
    """
    Generate a partial regression plot from the best-fit line
    of the predictor `pred`, where the intercept is calculated
    from the average of the `other` predictors,
    including the value of categorical predictors `cats` in the slope.
    """
    ax = plt.gca() if ax is None else ax
    data = lmfit.model.data.orig_exog
    params = lmfit.params
    allpreds = set(data.columns) - {"Intercept"}
    allnoncatpreds = set([pred for pred in allpreds if "T." not in pred])
    assert pred in allnoncatpreds
    others = allnoncatpreds - {pred} if others is None else others
    intercept = params["Intercept"]
    for other in others:
        intercept += params[other]*data[other].mean() 
    for cat in cats:
        intercept += params[cat]
    slope = params[pred]
    print(pred, "intercept=", intercept, "slope=", slope)
    xs = np.linspace(data[pred].min(), data[pred].max())
    ys = intercept + slope*xs
    sns.lineplot(x=xs, y=ys, color=color, ax=ax, linestyle=linestyle)
    return ax



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import re
from scipy.stats import t as tdist
import seaborn as sns
import statsmodels.formula.api as smf
from statsmodels.nonparametric import smoothers_lowess



# Simple linear regression
################################################################################

def plot_reg(lmres, ax=None):
    """
    Plot the best-fitting line of the simple linear regression
    model `lmres` on top of a scatter plot of the original data.
    """
    ax = plt.gca() if ax is None else ax
    # Draw scatter plot of the data
    xs = lmres.model.data.orig_exog.iloc[:,1]
    ys = lmres.model.data.orig_endog.iloc[:,0]
    sns.scatterplot(x=xs, y=ys, ax=ax)
    # Plot the model's predictions
    xgrid = np.linspace(np.min(xs), np.max(xs), 100)
    ypred = lmres.get_prediction({xs.name: xgrid})
    sns.lineplot(x=xgrid, y=ypred.predicted, ax=ax)
    return ax


def plot_resid(lmres, pred=None, lowess=False, ax=None):
    """
    Residuals plot for the model `lmres` vs. the predictor `pred`.
    If `pred` is None, plot the residuals vs. the fitted values.
    The function can an optional LOWESS curve for the residuals.
    """
    ax = plt.gca() if ax is None else ax
    if pred is None or pred in ["fittedvalues", "predictions"]:
        xs = lmres.fittedvalues
        xname = "fitted values"
    else:
        xs = lmres.model.data.orig_exog[pred]
        xname = pred
    ys = lmres.resid
    sns.scatterplot(x=xs, y=ys, ax=ax)
    ax.axhline(y=0, color="k", linestyle="dotted", alpha=0.6)
    if lowess:
        xgrid, ylowess = smoothers_lowess.lowess(ys, xs, frac=0.72).T
        sns.lineplot(x=xgrid, y=ylowess, ax=ax)
    ax.set_xlabel(xname)
    ax.set_ylabel("residuals $r_i$")
    return ax


def plot_pred_bands(lmres, ax=None,
                    ci_mean=False, alpha_mean=0.1, lab_mean=True,
                    ci_obs=False, alpha_obs=0.1, lab_obs=True):
    """
    Plot the confidence intervals for the model predcitions.
    If `ci_mean` is True: draw a (1-alpha_mean)-CI for the mean.
    If `ci_obs` is True: draw a (1-ci_obs)-CI for the predicted values.
    """
    ax = plt.gca() if ax is None else ax
    xs = lmres.model.data.orig_exog.iloc[:,1]
    n = lmres.nobs

    # Get model predicitons
    xgrid = np.linspace(np.min(xs), np.max(xs), 100)
    ypred = lmres.get_prediction({xs.name: xgrid})

    if ci_mean:
        # Draw the confidence interval for the mean predictions
        t_05, t_95 = tdist(df=n-2).ppf([alpha_mean/2, 1-alpha_mean/2])
        lower_mean = ypred.predicted + t_05*ypred.se_mean
        upper_mean = ypred.predicted + t_95*ypred.se_mean
        if lab_mean:
            perc_mean = round(100*(1-alpha_mean))
            label_mean = f"{perc_mean}% confidence interval for the mean"
        else:
            label_mean = None
        ax.fill_between(xgrid, lower_mean, upper_mean, color="C0", alpha=0.4, label=label_mean)

    if ci_obs:
        # Draw the confidence interval for the predicted observations
        t_05, t_95 = tdist(df=n-2).ppf([alpha_obs/2, 1-alpha_obs/2])
        lower_obs = ypred.predicted + t_05*ypred.se_obs
        upper_obs = ypred.predicted + t_95*ypred.se_obs
        if lab_obs:
            perc_obs = round(100*(1-alpha_obs))
            label_obs = f"{perc_obs}% confidence interval for observations"
        else:
            label_obs = None
        ax.fill_between(xgrid, lower_obs, upper_obs, color="C0", alpha=0.1, label=label_obs)

    if (ci_mean and lab_mean) or (ci_obs and lab_obs):
        ax.legend()
    return ax



# Multiple linear regression
################################################################################

def plot_partreg(lmres, pred, ax=None):
    """
    Generate a partial regression plot from the linear model `lmres`
    against the predictor `pred`, given the other predictors.
    We plot the residuals of the model `outcome ~ other` along the y-axis,
    versus the residuals of the model `pred ~ other` along the x-axis.
    """
    ax = plt.gca() if ax is None else ax
    xdata = lmres.model.data.orig_exog
    ydata = lmres.model.data.orig_endog
    data = pd.concat([xdata, ydata], axis=1)

    # Send to specialized function if the model contains categorical predictors
    if any(["C(" in name for name in xdata.columns]):
        return plot_partreg_cat(lmres, pred, ax=ax)

    # Find others= as list of strings
    allpreds = list(xdata.columns)
    names_to_skip = ["Intercept", pred]
    others = [name for name in allpreds if name not in names_to_skip]
    others_formula = "1"
    if others:
        others_formula += "+" + "+".join(others)

    # x-axis = residuals of the model `pred ~ 1 + others`
    lmpred = smf.ols(f"{pred} ~ {others_formula}", data=data).fit()
    xresids = lmpred.resid

    # y-axis = residuals of the model `outcome ~ 1 + others`
    outname = lmres.model.endog_names
    lmoutcome = smf.ols(f"{outname} ~ {others_formula}", data=data).fit()
    yresids = lmoutcome.resid

    # Draw a scatter plot of the y-residuals vs. the x-residuals
    sns.scatterplot(x=xresids, y=yresids, ax=ax)

    # Plot the best-fitting line
    dfresids = pd.DataFrame({"xresids": xresids, "yresids": yresids})
    lmresids = smf.ols("yresids ~ 0 + xresids", data=dfresids).fit()
    slope = lmresids.params.iloc[0]
    xgrid = np.linspace(xresids.min(), xresids.max(), 100)
    ys = slope * xgrid
    sns.lineplot(x=xgrid, y=ys, ax=ax)

    # Add descriptive labels
    if len(others_formula) > 20:
        others_formula = "other"
    ax.set_xlabel(f"$\\mathtt{{{pred}}}$~$\\mathtt{{{others_formula}}}$ residuals")
    ax.set_ylabel(f"$\\mathtt{{{outname}}}$~$\\mathtt{{{others_formula}}}$ residuals")
    return ax


def plot_projreg(lmres, pred, others=None, ax=None):
    """
    Generate a partial regression plot from the best-fit line
    of the predictor `pred`, where the intercept is calculated
    from the average of the `other` predictors.
    """
    ax = plt.gca() if ax is None else ax
    xdata = lmres.model.data.orig_exog
    xs = xdata[pred]
    ys = lmres.model.endog
    yname = lmres.model.endog_names
    sns.scatterplot(x=xs, y=ys, ax=ax)
    params = lmres.params
    allpreds = set(xdata.columns) - {"Intercept"}
    assert pred in allpreds 
    others = allpreds - {pred} if others is None else others
    intercept = params["Intercept"]
    for other in others:
        intercept += params[other]*xdata[other].mean() 
    slope = params[pred]
    print(pred, "intercept=", intercept, "slope=", slope)
    xgrid = np.linspace(xs.min(), xs.max())
    ypred = intercept + slope*xgrid
    sns.lineplot(x=xgrid, y=ypred, ax=ax)
    ax.set_xlabel(pred)
    ax.set_ylabel(yname)
    return ax


def plot_partreg_cat(lmres, pred, ax=None):
    """
    A version of `plot_partreg` that can handle categorical predictors.
    """
    ax = plt.gca() if ax is None else ax
    xdata = lmres.model.data.orig_exog
    ydata = lmres.model.data.orig_endog
    data = pd.concat([xdata, ydata], axis=1)

    # Find others= as list of strings
    allpreds = list(xdata.columns)
    names_to_skip = ["Intercept", pred]
    others = [name for name in allpreds if name not in names_to_skip]
    others_formula = "1"
    others_display = "1"
    if others:
        others_quoted = ["Q('" + other + "')" for other in others]
        others_formula += "+" + "+".join(others_quoted)
        for other in others:
            m = re.match("C\\((?P<varname>.*)\\).*", other)
            if m:
                varname = m.group("varname")
                other_clean = "C(" + varname + ")"
            else:
                other_clean = other
            if other_clean not in others_display:
                others_display += "+" + other_clean

    # x-axis = residuals of `pred ~ 1 + others`
    lmpred = smf.ols(f"{pred} ~ {others_formula}", data=data).fit()
    xresids = lmpred.resid

    # y-axis = residuals of `outcome ~ 1 + others`
    outname = lmres.model.endog_names
    lmoutcome = smf.ols(f"{outname} ~ {others_formula}", data=data).fit()
    yresids = lmoutcome.resid

    # scatter plot
    sns.scatterplot(x=xresids, y=yresids, ax=ax)
    xlims = ax.get_xlim()
    ylims = ax.get_ylim()

    # best-fit line between the residuals
    dfresids = pd.DataFrame({"xresids": xresids, "yresids": yresids})
    lmresids = smf.ols("yresids ~ 0 + xresids", data=dfresids).fit()
    slope = lmresids.params.iloc[0]
    xs = np.linspace(*ylims, 100)
    ys = slope*xs
    sns.lineplot(x=xs, y=ys, ax=ax)

    # ax.set_title('Partial regression plot')
    if len(others_display) > 20:
        others_display = "other"
    ax.set_xlabel(f"{pred} ~ {others_display}  residuals")
    ax.set_ylabel(f"{outname} ~ {others_display}  residuals")
    ax.set_xlim(xlims)
    ax.set_ylim(ylims)
    return ax



# Diagnostic plots
################################################################################

def plot_scaleloc(lmres, lowess=True):
    """
    Plot the scale-location plot for the linear model `lmres`.
    """
    sigmahat = np.sqrt(lmres.scale)
    std_resids = lmres.resid / sigmahat
    sqrt_abs_std_resids = np.sqrt(np.abs(std_resids))
    xs = lmres.fittedvalues
    ax = sns.regplot(x=xs, y=sqrt_abs_std_resids, lowess=True)
    # TODO: repalce with scatterplot + manual lowess lineplot
    ax.set_xlabel("fitted values")
    ax.set_ylabel(r"$\sqrt{|\mathrm{standardized residuals}|}$")
    return ax




from contextlib import contextmanager
import json
import logging
import matplotlib.pyplot as plt
import os
import shutil
import urllib.parse
import urllib.request



@contextmanager
def loglevel(level, module=None):
    """
    Context manager to set logging level locally.
    Useful for silencing the output of Bambi model fit method.
    """
    if isinstance(level, str):
        LEVEL_NAMES_MAPPING = {
            'CRITICAL': 50, 'FATAL': 50, 'ERROR': 40, 'WARN': 30,
            'WARNING': 30, 'INFO': 20, 'DEBUG': 10, 'NOTSET': 0
        }
        level = level.upper()
        level = LEVEL_NAMES_MAPPING[level]
    logger = logging.getLogger(module)
    current_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(current_level)


def ensure_containing_dir_exists(filepath):
    parent = os.path.join(filepath, os.pardir)
    absparent = os.path.abspath(parent)
    if not os.path.exists(absparent):
        os.makedirs(absparent)


def default_labeler(params, params_to_latex):
    """
    Returns string appropriate for probability distribution label used in plot.
    """
    DEFAULT_PARAMS_TO_LATEX = {
        'mu': '\\mu',
        'sigma': '\\sigma',
        'lambda': '\\lambda',
        'beta': '\\beta',
        'a': 'a',
        'b': 'b',
        'N': 'N',
        'K': 'K',
        'k': 'k',
        'n': 'n',
        'p': 'p',
        'r': 'r',
    }
    params_to_latex = dict(DEFAULT_PARAMS_TO_LATEX, **params_to_latex)
    label_parts = []
    for param, value in params.items():
        if param in params_to_latex:
            label_part = '$' + params_to_latex[param] + '=' + str(value) + '$'
        else:
            label_part = str(param) + '=' + str(value)
        label_parts.append(label_part)
    label = ', '.join(label_parts)
    return label


def savefigure(obj, filename, tight_layout_kwargs=None):
    """
    Save the figure associated with `obj` (axes or figure).
    Assumes `filename` is relative path to pdf to save to,
    e.g. `figures/stats/some_figure.pdf`.
    """
    ensure_containing_dir_exists(filename)
    if not filename.endswith(".pdf"):
        filename = filename + ".pdf"

    if isinstance(obj, plt.Axes):
        fig = obj.figure
    elif isinstance(obj, plt.Figure):
        fig = obj
    else:
        raise ValueError("First argument must be Matplotlib figure or axes")

    # remove surrounding whitespace as much as possible
    if tight_layout_kwargs:
        fig.tight_layout(**tight_layout_kwargs)
    else:
        fig.tight_layout()

    # save as PDF
    fig.savefig(filename, dpi=300, bbox_inches="tight", pad_inches=0)
    print("Saved figure to", filename)

    # save as PNG
    filename2 = filename.replace(".pdf", ".png")
    fig.savefig(filename2, dpi=300, bbox_inches="tight", pad_inches=0)
    print("Saved figure to", filename2)




# GITHUB DOWNLOAD HELPERS
################################################################################

def github_api_request(path, params=None):
    """
    Make a GET request to the GitHub REST API at the given path.
    No authentication (sufficient for occasional use).
    """
    base_url = "https://api.github.com"
    url = base_url + path
    if params:
        url += "?" + urllib.parse.urlencode(params)
    headers = {
        "User-Agent": "python-stdlib-client",
        "Accept": "application/vnd.github.v3+json",
    }
    req = urllib.request.Request(url, headers=headers)
    with urllib.request.urlopen(req) as resp:
        if resp.status != 200:
            raise RuntimeError(
                f"GitHub API error {resp.status}: {resp.read().decode(errors='replace')}"
            )
        return json.load(resp)


def list_dir_recursive(owner, repo, path, branch="main"):
    """
    Recursively list all files under `path` in the given repo+branch.
    Returns a list of file paths relative to the repo root.
    """
    api_path = f"/repos/{owner}/{repo}/contents/{path}"
    items = github_api_request(api_path, params={"ref": branch})
    files = []
    for item in items:
        item_type = item.get("type")
        item_path = item.get("path")
        if item_type == "file":
            files.append(item_path)
        elif item_type == "dir":
            files.extend(list_dir_recursive(owner, repo, item_path, branch=branch))
        # ignore symlinks, submodules etc.
    return files


def download_file_from_raw(owner, repo, branch, prefix, file_path, download_root):
    """
    Download file from GitHub and save to `download_root`.
    Remove the `prefix` from the repo path to avoid adding subdirectory.
    """
    # Compute relative local path
    prefix_slash = prefix + "/"
    if file_path.startswith(prefix_slash):
        relative = file_path[len(prefix_slash):]
    else:
        relative =  os.path.basename(file_path)
    local_path = os.path.join(download_root, relative)
    raw_url = (
        f"https://raw.githubusercontent.com/"
        f"{owner}/{repo}/{branch}/{file_path}"
    )
    # Ensure subdirectories exist
    os.makedirs(os.path.dirname(local_path), exist_ok=True)
    with urllib.request.urlopen(raw_url) as resp, open(local_path, "wb") as f:
        chunk = resp.read(8192)
        while chunk:
            f.write(chunk)
            chunk = resp.read(8192)
    return local_path


def download_files(owner, repo, branch, prefix, file_paths, download_root):
    """
    Download all files in `file_paths` to `download_root`,
    preserving their relative paths.
    """
    downloaded = []
    for fp in file_paths:
        local_path = download_file_from_raw(owner, repo, branch, prefix, fp, download_root)
        downloaded.append(local_path)
    return downloaded


def have_required_files(download_root, required_filenames):
    """
    Check if `download_root` exists and contains all `required_filenames`.
    """
    if not os.path.isdir(download_root):
        return False
    for name in required_filenames:
        path = os.path.join(download_root, name)
        if not os.path.isfile(path):
            return False
    return True




# DATASET AND SIMDATA DOWNLOAD HELPERS
################################################################################

OWNER = "minireference"
REPO = "noBSstats"
BRANCH = "main"


def ensure_datasets(verbose=False):
    """
    Make sure the `datasets/` folder with data files for
    the book is present in the current directory and.
    Otherwise:
    1. If ../datasets/ exists and contains the required files,
       copy from ../datasets/ to ./datasets/
    2. Download the files from GitHub repo:
       https://github.com/minireference/noBSstats/tree/main/datasets
    """
    DATASETS_PATH = "datasets"
    DATASETS_DOWNLOAD_ROOT = "datasets"
    PARENT_DATASETS = os.path.join("..", "datasets")
    REQUIRED_FILES = ["apples.csv", "doctors.csv", "eprices.csv"]    
    if have_required_files(DATASETS_DOWNLOAD_ROOT, REQUIRED_FILES):
        print(f"{DATASETS_DOWNLOAD_ROOT}/ directory already exists.")
    elif have_required_files(PARENT_DATASETS, REQUIRED_FILES):
        # Copy files from ../datasets/ to ./datasets/ (when working with .zip download)
        if verbose:
            print(f"Found valid datasets in {PARENT_DATASETS}/, copying recursively...")
        # Ensure local datasets folder exists
        os.makedirs(DATASETS_DOWNLOAD_ROOT, exist_ok=True)
        # Recursively copy contents of ../datasets/ into ./datasets/
        for root, dirs, files in os.walk(PARENT_DATASETS):
            # Compute relative path inside ../datasets/
            rel = os.path.relpath(root, PARENT_DATASETS)
            target_root = (DATASETS_DOWNLOAD_ROOT if rel == "." else os.path.join(DATASETS_DOWNLOAD_ROOT, rel) )
            os.makedirs(target_root, exist_ok=True)
            # Copy all files
            for fname in files:
                src = os.path.join(root, fname)
                dst = os.path.join(target_root, fname)
                shutil.copy2(src, dst)
                if verbose:
                    print(f"  copied {src} -> {dst}")
        print(f"Found {PARENT_DATASETS}/ and copied files to {DATASETS_DOWNLOAD_ROOT}/.")
    else:
        # Download files from GitHub
        all_files = list_dir_recursive(OWNER, REPO, DATASETS_PATH, branch=BRANCH)
        if verbose:
            print("Data files found in noBSstats repo:")
            for f in all_files:
                print("  ", f)
        downloaded_paths = download_files(
            OWNER, REPO, BRANCH, DATASETS_PATH, all_files, DATASETS_DOWNLOAD_ROOT
        )
        if verbose:
            print("\nDownloaded files:")
            for p in downloaded_paths:
                print("  ", p)
        print(f"Downloaded data files to {DATASETS_DOWNLOAD_ROOT}/ from GitHub.")


def ensure_simdata(force=False, verbose=False):
    """
    Download the simulation data used in the notebooks from:
    https://github.com/minireference/noBSstats/tree/main/notebooks/simdata
    Files are downloaded into the local directory:
        simdata/
    and flattened so that only the filenames (no subdirectories) are kept.
    """
    SIMDATA_PATH = "notebooks/simdata"  # directory within the repo
    SIMDATA_DOWNLOAD_ROOT = "simdata"        # local directory for simdata
    # Only hit the API if forced or the local directory doesn't exist yet
    if force or not os.path.isdir(SIMDATA_DOWNLOAD_ROOT):
        # List all files in notebooks/simdata
        all_files = list_dir_recursive(OWNER, REPO, SIMDATA_PATH, branch=BRANCH)
        if verbose:
            print("Simdata files found in noBSstats repo:")
            for f in all_files:
                print("  ", f)
        # Download files, flattening under SIMDATA_PATH
        downloaded_paths = download_files(
            OWNER, REPO, BRANCH, SIMDATA_PATH, all_files, SIMDATA_DOWNLOAD_ROOT
        )
        if verbose:
            print("\nDownloaded simdata files:")
            for p in downloaded_paths:
                print("  ", p)
    else:
        if verbose:
            print(f"{SIMDATA_DOWNLOAD_ROOT}/ already exists; skipping download.")
    print(f"{SIMDATA_DOWNLOAD_ROOT}/ directory present and ready.")




import math
from scipy.stats import expon
from scipy.stats import ks_1samp

# SUT
from ministats.formulas import ks_distance



def test_ks_distance_against_scipy():
    rvE = expon(loc=0, scale=5)
    exp_data = rvE.rvs(10)
    ks_dist = ks_distance(exp_data, rvE)
    ks_dist_scipy = ks_1samp(exp_data, rvE.cdf).statistic
    assert math.isclose(ks_dist, ks_dist_scipy)



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..utils import savefigure


# Discrete random variables
################################################################################

def plot_pmf(rv, xlims=None, ylims=None, rv_name="X", ax=None, title=None, label=None):
    """
    Plot the PMF of the discrete random variable `rv` over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin = 0
        smax = rv.support()[1]
        xmax = rv.ppf(0.99999) if smax == np.inf else smax + 1
    xs = np.arange(xmin, xmax)

    # Compute the probability mass function and plot it
    fXs = rv.pmf(xs)
    fXs = np.where(fXs == 0, np.nan, fXs)  # set zero fXs to np.nan
    ax.stem(xs, fXs, basefmt=" ", label=label)
    ax.set_xticks(xs)
    ax.set_xlabel("$" + rv_name.lower() + "$")
    ax.set_ylabel(f"$f_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    if label:
        ax.legend()

    if title and title.lower() == "auto":
        title = "Probability mass function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax


def plot_cdf(rv, xlims=None, ylims=None, rv_name="X", ax=None, title=None, **kwargs):
    """
    Plot the CDF of the random variable `rv` (discrete or continuous) over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin, xmax = rv.ppf(0.000000001), rv.ppf(0.99999)
    xs = np.linspace(xmin, xmax, 1000)

    # Compute the CDF and plot it
    FXs = rv.cdf(xs)
    sns.lineplot(x=xs, y=FXs, ax=ax, **kwargs)

    # Set plot attributes
    ax.set_xlabel("$b$")
    ax.set_ylabel(f"$F_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    if title and title.lower() == "auto":
        title = "Cumulative distribution function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax



# Continuous random variables
################################################################################

def plot_pdf(rv, xlims=None, ylims=None, rv_name="X", a=None, b=None, ax=None, title=None, **kwargs):
    """
    Plot the PDF of the continuous random variable `rv` over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin, xmax = rv.ppf(0.000000001), rv.ppf(0.99999)
    xs = np.linspace(xmin, xmax, 1000)

    # Compute the probability density function and plot it
    fXs = rv.pdf(xs)
    sns.lineplot(x=xs, y=fXs, ax=ax, **kwargs)
    ax.set_xlabel("$" + rv_name.lower() + "$")
    ax.set_ylabel(f"$f_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    
    if a or b:
        # Highlight the area under fX between x=a and x=b
        if a is None:
            a = rv.support()[0]
        if b is None:
            b = rv.support()[1]

        mask = (xs > a) & (xs < b)
        ax.fill_between(xs[mask], y1=fXs[mask], alpha=0.2)
        ax.vlines([a], ymin=0, ymax=rv.pdf(a), linestyle="-", alpha=0.5)
        ax.vlines([b], ymin=0, ymax=rv.pdf(b), linestyle="-", alpha=0.5)

    if title and title.lower() == "auto":
        title = "Probability density function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax



# Joint distribution plots
################################################################################

def get_meshgrid_and_pos(xlims, ylims, ngrid):
    """
    Create two 1D grids with `ngrid` points in each dimension,
    then combine them using meshgrid and stack the results along a third dimension
    as required to evaluate multivariate probability density function.
    """
    xmin, xmax = xlims
    ymin, ymax = ylims
    xs = np.linspace(xmin, xmax, ngrid)
    ys = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(xs, ys)
    pos = np.empty(X.shape + (2,))
    pos[:, :, 0] = X
    pos[:, :, 1] = Y
    # ALT.
    # pos = np.dstack( (X, Y) )
    return X, Y, pos


def plot_joint_pdf_contourf(rvXY, xlims, ylims, ngrid=200, ax=None):
    """
    Contour plot of a bivariate joint distribution `rvXY`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots(figsize=(7,4))
    else:
        fig = ax.figure

    # Compute the joint-probability density function values
    X, Y, pos = get_meshgrid_and_pos(xlims, ylims, ngrid)
    fXY = rvXY.pdf(pos)

    # Contour plot
    cax = ax.contourf(fXY,
                      origin='lower',
                      extent=(*xlims, *ylims),
                      levels=10,
                      cmap="Greys")
    ax.set_xlabel('$x$')
    ax.set_ylabel('$y$')

    return ax


def plot_joint_pdf_surface(rvXY, xlims, ylims, ngrid=200, fig=None, viewdict=None):
    """
    Surface plot of a bivariate joint distribution `rvXY`.
    https://stackoverflow.com/questions/38698277/plot-normal-distribution-in-3d
    """
    # Setup figure and axes
    if fig is None:
        fig = plt.figure(figsize=(7,7))
    ax = plt.axes(projection='3d')

    # Compute the joint-probability density function values
    X, Y, pos = get_meshgrid_and_pos(xlims, ylims, ngrid)
    fXY = rvXY.pdf(pos)

    # Generate the 3D surface plot
    ax.plot_surface(X, Y, fXY, cmap='Greys', linewidth=0)
    ax.set_box_aspect((xlims[1]-xlims[0], ylims[1]-ylims[0], 3))
    if viewdict is not None:
        ax.view_init(**viewdict)
    ax.set_xlabel('$x$')
    ax.set_ylabel('$y$')
    ax.set_zlabel('$f_{XY}$')
    ax.set_zticks([])

    return ax




# Diagnostic plots (used in Section 2.7 Random variable generation)
################################################################################


def plot_epmf(data, xlims=None, ylims=None, name="xs", ax=None, title=None, label=None):
    """
    Plot the empirical pmf of the observations in  `data`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots(figsize=(5,1.2))
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        extrax = 0.3 * max(data)  # extend further to the right
        xmin = int(min(data))
        xmax = int(max(data) + extrax)

    # Compute the probability mass function and plot it
    data = np.array(data)
    n = len(data)
    xs, counts = np.unique(data, return_counts=True)
    fxs = counts / n

    label = f"epmf({name})"
    ax.stem(xs, fxs, basefmt=" ", label=label)
    ax.set_xticks(range(xmin,xmax))
    # ax.set_ylim([-0.01,0.22])
    # ax.set_yticks([0, 0.1, 0.2])
    ax.set_xlabel("$b$")
    ax.set_ylabel(f"$f_{{\\text{{{name}}}}}$")
    # ax.set_xticks(xs)
    # ax.set_xlabel(rv_name.lower())
    # ax.set_ylabel(f"$f_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    if label:
        ax.legend()

    if title and title.lower() == "auto":
        title = "Empirical probability mass function of the data " + name
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax


def plot_ecdf(data, xlims=None, ylims=None, name="xs", ax=None, title=None, label=None):
    """
    Plot the empirical CDF of the observations in  `data`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots(figsize=(5,2))
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        extrax = 0.3 * max(data)  # extend further to the right
        xmin = int(min(data))
        xmax = int(max(data) + extrax)
    

    def ecdf(data, b):
        sorted_data = np.sort(data)
        count = sum(sorted_data <= b)  # num. of obs. <= b
        return count / len(data)       # proportion of total

    # Compute the probability mass function and plot it
    data = np.array(data)
    bs = np.linspace(0, xmax, 1000)
    Fxs = [ecdf(data,b) for b in bs]
    # label = f"eCDF({name})"
    ax = sns.lineplot(x=bs, y=Fxs, drawstyle='steps-post', label=label)
    ax.set_xlabel("$b$")
    ax.set_ylabel(f"$F_{{\\text{{{name}}}}}$")
    ax.set_xlim([0, xmax])
    ax.set_xticks(range(0,xmax))
    if ylims:
        ax.set_ylim(*ylims)
    if label:
        ax.legend()

    if title and title.lower() == "auto":
        title = "Empirical cumulative distribution function of the data " + name
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax





def qq_plot(data, dist, ax=None, xlims=None, filename=None, **kwargs):
    """
    This function qq_plot tries to imitate the behaviour of the function `qqplot`
    defined in `statsmodels.graphics.api`. Usage: `qqplot(data, dist=norm(0,1), line='q')`. See:
    https://github.com/statsmodels/statsmodels/blob/main/statsmodels/graphics/gofplots.py#L912-L919
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Add the Q-Q scatter plot
    n = len(data)
    qs = np.linspace(1/(n+1), n/(n+1), n)
    xs = dist.ppf(qs)
    sorted_data = np.sort(data)
    ys = sorted_data
    # ALT. ys = np.quantile(data, qs, method="inverted_cdf")
    sns.scatterplot(x=xs, y=ys, ax=ax, alpha=0.7, **kwargs)

    # Compute the parameters m and b for the diagonal line
    xq25, xq75 = dist.ppf([0.25, 0.75])
    yq25, yq75 = np.quantile(data, [0.25, 0.75])
    m = (yq75 - yq25) / (xq75 - xq25)
    b = yq25 - m * xq25
    # add the line  y = m*x+b  to the plot
    linexs = np.linspace(min(xs), max(xs))
    lineys = m * linexs + b
    sns.lineplot(x=linexs, y=lineys, ax=ax, color="r")

    # Handle keyword arguments
    if xlims:
        ax.set_xlim(xlims)
    if filename:
        savefigure(ax, filename)

    return ax



import numpy as np

# SAMPLING DISTRIBUTIONS
################################################################################

def gen_sampling_dist_of_mean(rvX, n, N=1000):
    """
    Generate the sampling distribution of the mean for samples of size `n`
    from the random variable `rvX` based on `N` simulated random samples.
    """
    xbars = []
    for j in range(N):
        xsample = rvX.rvs(n)
        xbar = np.mean(xsample)
        xbars.append(xbar)
    return xbars


# This function is shown in Section 2.8
# The re-definition is identical, but uses `estfunc` instead of `statfunc`
def gen_sampling_dist(rvX, statfunc, n, N=1000):
    """
    Simulate `N` samples of size `n` from the random variable `rvX`
    to generate the sampling distribution of the statistic `statfunc`.
    """
    stats = []
    for j in range(N):
        xsample = rvX.rvs(n)
        stat = statfunc(xsample)
        stats.append(stat)
    return stats



def gen_sampling_dist(rv, estfunc, n, N=10000):
    """
    Simulate `N` samples of size `n` from the random variable `rv`
    to generate the sampling distribution of the estimator `estfunc`.
    """
    estimates = []
    for j in range(N):
        sample = rv.rvs(n)
        estimate = estfunc(sample)
        estimates.append(estimate)
    return estimates




# BOOTSTRAP
################################################################################

def gen_boot_dist(sample, estfunc, B=5000):
    """
    Generate estimates from the sampling distribution of the estimator `estfunc`
    based on `B` bootstrap samples (sampling with replacement) from `sample`.
    """
    n = len(sample)
    bestimates = []
    for j in range(B):
        bsample = np.random.choice(sample, n, replace=True)
        bestimate = estfunc(bsample)
        bestimates.append(bestimate)
    return bestimates




import numpy as np
import statsmodels.api as sm


def calc_lm_vif(lmres, pred):
    """
    Calculate the variance inflation factor of the `pred` (str)
    for the linear model fit `lmfit`.
    """
    dmatrix = lmres.model.exog
    pred_idx = lmres.model.exog_names.index(pred)
    n_cols = dmatrix.shape[1]
    x_i = dmatrix[:, pred_idx]
    mask = np.arange(n_cols) != pred_idx
    X_noti = dmatrix[:, mask]
    r_squared_i = sm.OLS(x_i, X_noti).fit().rsquared
    vif = 1. / (1. - r_squared_i)
    return vif





import copy
import logging

import arviz as az
import bambi as bmb
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import fmin
from scipy.stats import gaussian_kde

# Silence annoying warning about missing BLAS library
# WARNING: Using NumPy C-API based implementation for BLAS functions.
logging.getLogger("pytensor.tensor.blas").setLevel(logging.ERROR)


    
# ESTIMATORS
################################################################################

def mode_from_samples(samples, prec=0.0001, **kwargs):
    """
    Calculate the mode (maximum) of the empirical distribution of `samples` by
    approximating the probability density function using KDE and finding its max.
    See also `az.plots.plot_utils.calculate_point_estimate("mode", samples)`.
    """
    values = np.arange(min(samples), max(samples), prec)
    kde = gaussian_kde(samples, **kwargs)
    return values[np.argmax(kde(values))]


# CREDIBILITY INTERVALS
################################################################################

def hdi_from_grid(params, probs, hdi_prob=0.9):
    """
    Compute the highest density interval from the grid approximation specified
    as probabilities `probs` evaluated at grid values `params`.
    """
    assert len(params) == len(probs)
    cdfprobs = np.cumsum(probs)

    def ppf(prob):
        "Inverse CDF function calculated from `params` grid and `prob`."
        idx = cdfprobs.searchsorted(prob)
        return params[idx]

    def width(startprob):
        "Calculate the width of the `hdi_prob` interval starting at `startprob`."
        return ppf(startprob + hdi_prob) - ppf(startprob)

    idx_left_stop = cdfprobs.searchsorted(1 - hdi_prob)
    widths = []
    for idx_probstart in range(0, idx_left_stop):
        probstart = cdfprobs[idx_probstart]
        widths.append(width(probstart))

    idx_left = np.argmin(widths)
    hdi_left = params[idx_left]
    probstart = cdfprobs[idx_left]
    idx_right = cdfprobs.searchsorted(probstart + hdi_prob)
    hdi_right = params[idx_right]
    return [hdi_left, hdi_right]


def hdi_from_rv(rv, hdi_prob=0.9):
    probstart0 = (1 - hdi_prob) / 2
    def width(probstart):
        return rv.ppf(probstart + hdi_prob) - rv.ppf(probstart)
    min_probstart = fmin(width, x0=probstart0, xtol=1e-9, disp=False)[0]
    hdi_left = rv.ppf(min_probstart)
    hdi_right = rv.ppf(min_probstart + hdi_prob)
    return [hdi_left, hdi_right]


def hdi_from_samples(samples, hdi_prob=0.9):
    samples = np.sort(samples)
    n = len(samples)
    idx_width = int(np.floor(hdi_prob*n))
    n_intervals = n - idx_width
    left_endpoints = samples[0:n_intervals]
    right_endpoints = samples[idx_width:n]
    widths = right_endpoints - left_endpoints
    idx_left = np.argmin(widths)
    hdi_left = samples[idx_left]
    hdi_right = samples[idx_left + idx_width]
    return [hdi_left, hdi_right]


def hdi_from_idata(idata, var_name, hdi_prob=0.9):
    stats_df = az.summary(idata, kind="stats", hdi_prob=hdi_prob, var_names=var_name)
    hdi_left = stats_df.iloc[0,2]
    hdi_right = stats_df.iloc[0,3]
    return [hdi_left, hdi_right]




# COMPARING TWO GROUPS
################################################################################

def bayes_dmeans(xsample, ysample, priors=None,
                 var_name="var", group_name="group", groups=["x", "y"]):
    """
    Compare the means of two groups using a Bayesian model.
    Returns a tuple containing the Bambi model and the InferenceData object.

    Usage example:
    >>> treated = iqs2[iqs2["group"]=="treat"]["iq"].values
    >>> controls = iqs2[iqs2["group"]=="ctrl"]["iq"].values
    >>> mod, idata = bayes_dmeans(treated, controls, var_name="iq",
                                  group_name="group", groups=["treat", "ctrl"])
    """
    # Pacakge raw data samples as a DataFrame
    m, n = len(xsample), len(ysample)
    groups_col = [groups[0]]*m + [groups[1]]*n
    var_col = list(xsample) + list(ysample)
    df = pd.DataFrame({group_name:groups_col, var_name:var_col})

    # Build the Bambi model
    formula = bmb.Formula(f"{var_name} ~ 1 + {group_name}",
                          f"sigma ~ 0 + {group_name}")
    model = bmb.Model(formula=formula,
                      family="t",
                      link="identity",
                      priors=priors,
                      data=df)

    # Fit the model
    idata = model.fit(draws=2000)
    return model, idata


def _infer_groups_from_idata(idata, group_name="group", other="other"):
    """
    Helper function used in `calc_dmeans_stats` and `plot_dmeans_stats`.
    """
    post = idata["posterior"]
    sigma_group = "sigma_" + group_name
    group_dim = group_name + "_dim"
    if sigma_group in post.data_vars and "sigma" not in post.data_vars:
        # Infer `groups` from the `sigma_{group_name}_dim` coordinate values
        sigma_group_dim = "sigma_" + group_name + "_dim"
        groups = list(post.coords[sigma_group_dim].values)
    elif group_dim in post.coords and len(post.coords[group_dim]) == 2:
        # Infer `groups` from the `{group_name}_dim` coordinate values
        groups = list(post.coords[group_dim].values)
    else:
        # Fallback: infer one group name from the `{group_name}_dim` dimension,
        # and label the other one as generic name `other`
        known_group = list(post.coords[group_dim].values)[0]
        groups = [other, known_group]
    return groups


def calc_dmeans_stats(idata, group_name="group"):
    """
    Calculate derived quantities used for the analysis plots and summaries.
    Handles both cases where formula `y ~ 1 + group` or `y ~ 0 + group` is used,
    and either group-specific `sigma` or common `sigma`.
    """
    post = idata["posterior"]
    groups = _infer_groups_from_idata(idata, group_name=group_name)

    # Add aliases for individual means and calculate the difference between means
    group_dim = group_name + "_dim"
    if "Intercept" in post.data_vars:
        # CASE A: formula is specified as `y ~ 1 + group`
        post["dmeans"] = post[group_name].loc[{group_dim:groups[1]}]
        post["mu_" + groups[0]] = post["Intercept"]
        post["mu_" + groups[1]] = post["Intercept"] + post["dmeans"]
    else:
        # CASE B: formula is specified as `y ~ 0 + group`
        post["mu_" + groups[0]] = post[group_name].loc[{group_dim:groups[0]}]
        post["mu_" + groups[1]] = post[group_name].loc[{group_dim:groups[1]}]
        post["dmeans"] = post["mu_" + groups[1]] - post["mu_" + groups[0]]

    sigma_group = "sigma_" + group_name
    if sigma_group in post.data_vars and "sigma" not in post.data_vars:
        # Calculate sigmas from log-sigmas
        sigma_group_dim = "sigma_" + group_name + "_dim"
        log_sigma_x = post[sigma_group].loc[{sigma_group_dim:groups[0]}]
        log_sigma_y = post[sigma_group].loc[{sigma_group_dim:groups[1]}]
        sigma_x_name = "sigma_" + groups[0]
        sigma_y_name = "sigma_" + groups[1]
        post[sigma_x_name] = np.exp(log_sigma_x)
        post[sigma_y_name] = np.exp(log_sigma_y)
        # Calculate the difference between standard deviations
        post["dsigmas"] = post[sigma_y_name] - post[sigma_x_name]
        # Effect size
        var_pooled = (post[sigma_x_name]**2 + post[sigma_y_name]**2) / 2
        post["cohend"] = post["dmeans"] / np.sqrt(var_pooled)
    else:
        # post["sigma"] is already on the right scale
        # Effect size
        post["cohend"] = post["dmeans"] / post["sigma"]

    return idata



def plot_dmeans_stats(model, idata, group_name="group",
                      figsize=(8,10), ppc_xlims=None, ppc_ylims=None):
    """
    Generate posterior panel of plots similar to the one in BEST paper.

    When model has group-specific `sigma`, the plot will look like this:
    +---------+--------------+
    | mu1     | post pred 1  |
    | mu2     | post pred 2  |
    | sigma1  | dmeans       |
    | sigma2  | dsigmas      |
    | nu      | cohend       |
    +---------+--------------+

    For analyzes with a common `sigma`, the plot will look like this:
    +---------+--------------+
    | mu1     | post pred 1  |
    | mu2     | post pred 2  |
    | dmeans  | dsigmas      |
    | nu      | cohend       |
    +---------+--------------+
    """
    post = idata["posterior"]
    groups = _infer_groups_from_idata(idata, group_name=group_name)

    # Compute posterior predictive checks 
    N_rep = 20
    draws_subset = np.random.choice(post["draw"].values, N_rep, replace=False)
    idata_rep = idata.sel(draw=draws_subset)
    df = model.data
    # PPC group 1
    data1 = df[df[group_name]==groups[1]]
    idata_rep1 = copy.deepcopy(idata_rep).sel(__obs__=data1.index)
    model.predict(idata_rep1, data=data1, kind="response")
    # PPC group 0
    if groups[0] == "other":
        altgroups = list(model.data[group_name].unique())
        altgroups.remove(groups[1])
        data0 = df[df[group_name]==altgroups[0]]
    else:
        data0 = df[df[group_name]==groups[0]]
    idata_rep0 = copy.deepcopy(idata_rep).sel(__obs__=data0.index)
    model.predict(idata_rep0, data=data0, kind="response")
    # Set x-lims automatically based on range of outcome variable
    if ppc_xlims is None:
        oname = model.response_component.term.name
        obs = df[oname]
        omin, omax, orange = obs.min(), obs.max(), obs.max()-obs.min()
        ppc_xlims = [omin-0.1*orange, omax+0.1*orange]

    with plt.rc_context({"figure.figsize":figsize}):

        sigma_group = "sigma_" + group_name
        if sigma_group in post.data_vars and "sigma" not in post.data_vars:
            fig, axs = plt.subplots(5,2)
            axmu1, axpp1        = axs[0,0], axs[0,1]
            axmu2, axpp2        = axs[1,0], axs[1,1]
            axsigma1, axdmeans  = axs[2,0], axs[2,1]
            axsigma2, axdsigmas = axs[3,0], axs[3,1]
            axnu, axcohend      = axs[4,0], axs[4,1]
        else:
            fig, axs = plt.subplots(4,2)
            axmu1, axpp1       = axs[0,0], axs[0,1]
            axmu2, axpp2       = axs[1,0], axs[1,1]
            axdmeans, axsigma  = axs[2,0], axs[2,1]
            axnu, axcohend     = axs[3,0], axs[3,1]
 
        # Top
        ## Left column
        az.plot_posterior(idata, group="posterior", var_names=["mu_" + groups[0]],
                          round_to=3, ax=axmu1)
        az.plot_posterior(idata, group="posterior", var_names=["mu_" + groups[1]],
                          round_to=3, ax=axmu2)
        ## Right column
        az.plot_ppc(idata_rep0, group="posterior", mean=False, ax=axpp1)
        axpp1.set_xlim(ppc_xlims)
        axpp1.set_xlabel(None)
        axpp1.set_title("Posterior predictive for " + groups[0])
        az.plot_ppc(idata_rep1, group="posterior", mean=False, ax=axpp2)
        axpp2.set_xlim(ppc_xlims)
        axpp2.set_xlabel(None)
        axpp2.set_title("Posterior predictive for " + groups[1])
        # Set same y-lims for the two PPC plots
        if ppc_ylims is None:
            ymin = min(axpp1.get_ylim()[0], axpp2.get_ylim()[0])
            ymax = max(axpp1.get_ylim()[1], axpp2.get_ylim()[1])
            axpp1.set_ylim([ymin,ymax])
            axpp2.set_ylim([ymin,ymax])
        else:
            axpp1.set_ylim(ppc_ylims)
            axpp2.set_ylim(ppc_ylims)

        # Middle
        if sigma_group in post.data_vars and "sigma" not in post.data_vars:
            az.plot_posterior(idata, group="posterior", var_names=["sigma_" + groups[0]],
                              point_estimate="mode", round_to=3, ax=axsigma1)
            az.plot_posterior(idata, group="posterior", var_names=["sigma_" + groups[1]],
                              point_estimate="mode", round_to=3, ax=axsigma2)
            az.plot_posterior(idata, group="posterior", var_names=["dsigmas"], hdi_prob=0.95,
                              point_estimate="mode", round_to=3, ax=axdsigmas)
            axdsigmas.axvline(0, c="C2", lw=2)
        else:
            az.plot_posterior(idata, group="posterior", var_names=["sigma"], hdi_prob=0.95,
                              point_estimate="mode", ax=axsigma)
            axsigma.axvline(0, c="C2", lw=2)
        az.plot_posterior(idata, group="posterior", var_names=["dmeans"], hdi_prob=0.95,
                          round_to=3, ax=axdmeans)
        axdmeans.axvline(0, c="C2", lw=2)

        # Bottom
        az.plot_posterior(idata, group="posterior", var_names=["nu"],
                          point_estimate="mode", ax=axnu)
        az.plot_posterior(idata, group="posterior", var_names=["cohend"],
                          point_estimate="mode", round_to=3, ax=axcohend)

    fig.tight_layout()
    return fig




# Bayesian Estimation Supersedes the t-Test (BEST) priors
################################################################################

def best_dmeans_model(xsample, ysample, nuprior="exp"):
    """
    Fit the model described in the "Bayesian Estimation Supersedes the t-Test"
    paper by John K. Kruschke.
    The function supports three different choices for the priors on `nu`:
      - `shiftedexp` = Expon(lam=1/29) + 1: the prior from the original paper
      - `exp` = Expon(lam=1/30): a simplified version without the +1 shift
      - `gamma` = Gamma(alpha=2.0, beta=0.1): the Bambi default prior for `nu`s
    Returns the Bambi model, which you can then analyze and fit.
    """
    pass


def best_dmeans_calc(idata, var_name="z", group_names=["treatment", "control"]):
    """
    Performs various calculations on the inference data object `idata`:
      - `dmeans`: difference between groups means
      - `dsigmas`: difference between groups standard deviaitons
      - 
      - `log10(nu)`: the normality parameter 

    """
    pass


def best_dmeans_plots():
    """
    Generte the panel of plots similar to the BEST paper.
    """
    pass



# BAYES FACTORS
################################################################################

# MAYBE: import grid approximaiotn methods from 50_extra_bayesian_stuff.ipynb




import numpy as np

from .estimators import mean
from .estimators import var


# STANDARDIZED EFFECT SIZE MEASURES
################################################################################

def cohend(sample, mu0):
    """
    Compute Cohen's d for one group compared to the theoretical mean `mu0`.
    """
    mean = np.mean(sample)
    std = np.std(sample, ddof=1)
    d = (mean - mu0) / std
    return d


def cohend2(sample1, sample2):
    """
    Compute Cohen's d measure of effect size for two independent samples.
    """
    n1, n2 = len(sample1), len(sample2)
    mean1, mean2 = mean(sample1), mean(sample2)
    var1, var2 = var(sample1), var(sample2)
    # calculate the pooled variance and standard deviation
    varp = ((n1-1)*var1 + (n2-1)*var2) / (n1 + n2 - 2)
    stdp = np.sqrt(varp)
    d = (mean1 - mean2) / stdp
    return d



# UTILS
################################################################################

def calcdf(stdX, n, stdY, m):
    """
    The Welch-Satterthwaite formula for the degrees of freedom parameter
    used for confidence intervals and Welch's t-test.
    """
    vX = stdX**2 / n
    vY = stdY**2 / m
    df = (vX + vY)**2 / (vX**2/(n-1) + vY**2/(m-1))
    return df



# KOLMOGOROV-SMIRNOV DISTANCE
################################################################################

def ks_distance(sample, rv):
    """
    Compute the Kolmogorov-Smirnov distance between an data in `sample` and 
    and the CDF function of the random variable `rv`.
    """
    n = len(sample)
    sample = np.sort(sample)

    # Empirical CDF values: i/n for the right-limit, (i-1)/n for left-limit
    ecdf_vals = np.arange(1, n+1) / n
    cdf_vals = rv.cdf(sample)

    # Compute KS distances
    ks_dist_plus = np.max(ecdf_vals - cdf_vals)
    ks_dist_minus = np.max(cdf_vals - (np.arange(0, n) / n))
    max_ks_dist = max(ks_dist_plus, ks_dist_minus)

    return max_ks_dist



import numpy as np
from scipy.stats import chi2
from scipy.stats import t as tdist

from .estimators import mean
from .estimators import var
from .formulas import calcdf
from .sampling import gen_boot_dist




# CONFIDENCE INTERVALS
################################################################################

def ci_mean(sample, alpha=0.1, method="a"):
    """
    Compute confidence interval for the population mean.
    - method="a" analytical approx. based on Student's t-dist
    - method="b" approx. based on bootstrap estimation
    """
    assert method in ["a", "b"]
    if method == "a":        # analytical approximation
        from scipy.stats import t as tdist
        n = len(sample)
        xbar = np.mean(sample)
        sehat = np.std(sample, ddof=1) / np.sqrt(n)
        t_l = tdist(df=n-1).ppf(alpha/2)
        t_u = tdist(df=n-1).ppf(1-alpha/2)
        return [xbar + t_l*sehat, xbar + t_u*sehat]
    elif method == "b":      # bootstrap estimation
        xbars_boot = gen_boot_dist(sample, estfunc=mean)
        return [np.quantile(xbars_boot, alpha/2),
                np.quantile(xbars_boot, 1-alpha/2)]


def ci_var(sample, alpha=0.1, method="a"):
    """
    Compute confidence interval for the population variance.
    - method="a" analytical approx. based on chi-square dist
    - method="b" approx. based on bootstrap estimation
    """
    assert method in ["a", "b"]
    if method == "a":        # analytical approximation
        n = len(sample)
        s2 = np.var(sample, ddof=1)
        q_l = chi2(df=n-1).ppf(alpha/2)
        q_u = chi2(df=n-1).ppf(1-alpha/2)
        return [(n-1)*s2/q_u, (n-1)*s2/q_l]
    elif method == "b":      # bootstrap estimation
        vars_boot = gen_boot_dist(sample, estfunc=var)
        return [np.quantile(vars_boot, alpha/2),
                np.quantile(vars_boot, 1-alpha/2)]


def ci_dmeans(xsample, ysample, alpha=0.1, method="a"):
    """
    Compute confidence interval for the difference between population means.
    - method="a" analytical approx. based on Student's t-dist
    - method="b" approx. based on bootstrap estimation
    """
    assert method in ["a", "b"]
    if method == "a":        # analytical approximation
        stdX, n = np.std(xsample, ddof=1), len(xsample)
        stdY, m = np.std(ysample, ddof=1), len(ysample)
        dhat = np.mean(xsample) - np.mean(ysample)
        seD = np.sqrt(stdX**2/n + stdY**2/m)
        dfD = calcdf(stdX, n, stdY, m)
        t_l = tdist(df=dfD).ppf(alpha/2)
        t_u = tdist(df=dfD).ppf(1-alpha/2)
        return [dhat + t_l*seD, dhat + t_u*seD]
    elif method == "b":      # bootstrap estimation
        xbars_boot = gen_boot_dist(xsample, np.mean)
        ybars_boot = gen_boot_dist(ysample, np.mean)
        dmeans_boot = np.subtract(xbars_boot, ybars_boot)
        return [np.quantile(dmeans_boot, alpha/2),
                np.quantile(dmeans_boot, 1-alpha/2)]




#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()


requirements = [
    "matplotlib>=3.7",
    "numpy>=1.24",
    "scipy>=1.6",
    "pandas>=2",
    "seaborn>=0.13.2",
    "statsmodels>=0.14.1",
    "bambi==0.15.0",    # pinned because of https://github.com/pymc-devs/pymc/issues/7978
    "pymc==5.23.0",     # pinned becasue of https://github.com/pymc-devs/pymc/issues/7978
    "arviz",
    "pingouin>=0.5.5",
]

test_requirements = ['pytest>=3', ]

setup(
    author="Ivan Savov",
    author_email='ivan@minireference.com',
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
    ],
    description="Common statistical testing procedures used for STATS 101 topics. The code is intentionally simple to make it easy to follow for beginners.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='ministats',
    name='ministats',
    packages=find_packages(include=['ministats', 'ministats.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/minireference/ministats',
    version='0.5.3',
    zip_safe=False,
)



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


# PLOT HELPERS
################################################################################

def plot_func(f, xlim=[0,5], flabel="f", ax=None):
    """
    Plot the function `f` over the interval `xlim`.
    """
    xs = np.linspace(xlim[0], xlim[1], 10000)
    fxs = fxs = np.array([f(x) for x in xs])
    ax = sns.lineplot(x=xs, y=fxs, ax=ax)    
    ax.set_xlim(*xlim)
    ax.set_xlabel("$x$")
    ax.set_ylabel(f"${flabel}(x)$")
    return ax


def plot_seq(ak, start=0, stop=10, label="$a_k$", ax=None):
    """
    Plot the sequence `ak` for between `start` and `stop`.
    """
    if ax is None:
        _, ax = plt.subplots()
    ks = np.arange(start, stop+1)
    aks = [ak(k) for k in ks]
    ax.stem(ks, aks, basefmt=" ")
    ax.set_xticks(ks)
    ax.set_xlabel("$k$")
    ax.set_ylabel(label)
    return ax





# CALCULUS OPERATIONS
################################################################################

def differentiate(f, x, delta=1e-9):
    """
    Compute the derivative of the function `f` at `x`
    using the rise-over-run calculation for run `delta`.
    """
    df = f(x+delta) - f(x)
    dx = delta
    return df / dx


def integrate(f, a, b, n=10000):
    """
    Compute the area under the graph of `f`
    between `x=a` and `x=b` using `n` rectangles.
    """
    dx = (b - a) / n                       # width of rectangular strips
    xs = [a + k*dx for k in range(1,n+1)]  # right-corners of the strips
    fxs = [f(x) for x in xs]               # heights of the strips
    area = sum([fx*dx for fx in fxs])      # total area
    return area



# CALCULUS PLOT HELPERS
################################################################################

def plot_integral(f, a=1, b=2, xlim=[0,5], flabel="f", ax=None, autolabel=False):
    """
    Plot the integral of `f` between `x=a` and `x=b`.
    """
    # Plot the function
    xs = np.linspace(xlim[0], xlim[1], 10000)
    fxs = fxs = np.array([f(x) for x in xs])
    ax = sns.lineplot(x=xs, y=fxs, ax=ax)    
    ax.set_xlim(*xlim)
    ax.set_xlabel("$x$")
    ax.set_ylabel(f"${flabel}(x)$")
    # Highlight the area under f(x) between x=a and x=b
    mask = (xs > a) & (xs < b)
    ax.fill_between(xs[mask], y1=fxs[mask], alpha=0.4)
    ax.vlines([a], ymin=0, ymax=f(a))
    ax.vlines([b], ymin=0, ymax=f(b))
    if autolabel:
        Alabel = f"$A_{{{flabel}}}({a},{b})$"
        ax.text((a+b)/2, 0.4*f((a+b)/2), Alabel, ha="center", fontsize="large");
    return ax


def plot_riemann_sum(f, a=1, b=2, xlim=[0,5], n=20, flabel="f", ax=None):
    """
    Draw the Riemann sum approximation to the integral of `f`
    between `x=a` and `x=b` using `n` rectangles.
    """
    # Calculate the value of the Riemann sum approximation
    dx = (b - a) / n                       # width of rectangular strips
    xs = [a + k*dx for k in range(1,n+1)]  # right-corners of the strips
    fxs = [f(x) for x in xs]               # heights of the strips
    area = sum([fx*dx for fx in fxs])      # total area
    print(f"Riemann sum with n={n} rectangles: approx. area ≈ {area:.5f}")
    # Plot the function
    xs_plot = np.linspace(xlim[0], xlim[1], 10000)
    fxs_plot = f(xs_plot)
    ax = sns.lineplot(x=xs_plot, y=fxs_plot, ax=ax)
    # Draw rectangles
    left_corners = [xr - dx for xr in xs]
    ax.bar(left_corners, fxs, width=dx, align="edge", edgecolor="black", alpha=0.3)
    ax.set_xlim(*xlim)
    ax.set_xlabel("$x$")
    ax.set_ylabel(f"${flabel}(x)$")    
    return ax


def plot_series(ak, start=0, stop=10, label="$a_k$", ax=None):
    """
    Draw a bar plot that corresponds to the series `sum(ak)`
    between `start` and `stop`.
    """
    if ax is None:
        _, ax = plt.subplots()
    # Plot the sequence
    ks = np.arange(start, stop+1)
    aks = [ak(k) for k in ks]
    ax.stem(ks, aks, basefmt=" ")
    # Compute the sum
    area = sum(aks)
    print(f"The sum of the first {stop-start+1} terms of the sequence is {area:.6f}")
    # Draw the series as rectangles
    ax.bar(ks, aks, width=1, align="edge", edgecolor="black", alpha=0.3)
    ax.set_xticks(ks)
    ax.set_xlabel("$k$")
    ax.set_ylabel(label)
    return ax



import math

from matplotlib import gridspec
from matplotlib.collections import PolyCollection
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats.contingency import margins
from scipy.stats import t as tdist
import seaborn as sns
import xarray as xr

from ..plots import nicebins
from ..plots.figures import calc_prob_and_plot
from ..plots.figures import calc_prob_and_plot_tails
from ..plots.probability import get_meshgrid_and_pos


# Probability theory
################################################################################

def plot_ks_dist_with_inset(sample, rv, label_sample="eCDF(sample)", label_rv="CDF $F_X$"):
    """
    Usage example:
    ```
    def gen_e(lam):
        u = np.random.rand()
        e = -1 * np.log(1-u) / lam
        return e
    np.random.seed(26)
    N = 200  # number of observations to generate
    es2 = [gen_e(lam=0.2) for i in range(0,N)]
    plot_ks_dist_with_inset(es2, rvE, label_sample="eCDF(es2)", label_rv="CDF $F_E$")
    ```
    """
    from matplotlib.patches import Rectangle
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes, mark_inset

    # KS distance function
    def ks_distance_and_location(sample, dist_cdf):
        sample_sorted = np.sort(sample)
        n = len(sample_sorted)
        ecdf_vals = np.arange(1, n+1) / n
        cdf_vals = dist_cdf(sample_sorted)
        diffs = np.abs(ecdf_vals - cdf_vals)
        D = np.max(diffs)
        idx = np.argmax(diffs)
        return D, sample_sorted[idx], ecdf_vals[idx], cdf_vals[idx]

    # Compute KS distance
    D, x_star, F_emp, F_th = ks_distance_and_location(sample, rv.cdf)

    # Plot
    fig, ax = plt.subplots(figsize=(8, 5))
    sns.ecdfplot(sample, label=label_sample, ax=ax)

    xrange = np.linspace(0, 30, 1000)
    sns.lineplot(x=xrange, y=rv.cdf(xrange), ax=ax, label=label_rv, color="C1")
    ax.legend()

    # Zoom range
    zoom_radius = 0.75
    x_min, x_max = x_star - zoom_radius, x_star + zoom_radius
    y_min, y_max = F_th - zoom_radius * 0.1, F_th + zoom_radius * 0.1

    # Add rectangle to show zoom region in main plot
    rect = Rectangle((x_min, y_min), x_max - x_min, y_max - y_min,
                     linewidth=1, edgecolor='black', facecolor='none', linestyle='--')
    ax.add_patch(rect)

    # Inset with 60% size
    axins = inset_axes(ax, width="40%", height="75%", loc="lower right")
    sns.ecdfplot(sample, ax=axins)
    sns.lineplot(x=xrange, y=rv.cdf(xrange), ax=axins, color="C1")

    # Inset limits
    axins.set_xlim(x_min, x_max)
    axins.set_ylim(y_min, y_max)
    axins.set_title("Zoom in near max $D_{KS}$", fontsize=11)

    # Draw short red line between empirical and theoretical CDF at x_star
    axins.plot([x_star, x_star], [F_emp, F_th], color='red', linestyle='--', linewidth=1.5)
    axins.annotate(
        f"$D_{{KS}} = {D:.4f}$",
        xy=(x_star, (F_emp + F_th)/2),             # Point to the middle of red line
        xytext=(x_star + 0.08, (F_emp + F_th)/2-0.05),   # Text placed to the right
        textcoords="data",
        arrowprops=dict(arrowstyle="->", color='red'),
        ha='left', va='center', fontsize=12, color='red'
    )

    # Draw connectors between inset and main plot
    mark_inset(ax, axins, loc1=2, loc2=3, fc="none", ec="0.5")

    # Return figure
    return fig



# Section 2.4: Multivariate distributions
################################################################################

def plot_joint_pdf_and_marginals(rvXY, xlims, ylims, ngrid=200, fig=None):
    """
    Contour plot of a bivariate joint distribution $f_XY$ (`rvXY.pdf`) while
    also showing the marginals $f_X$ and $f_Y$ on the sides.
    """
    # Setup figure and axes
    if fig is None:
        fig = plt.figure(figsize=(7,4))

    # Compute the joint-probability density function values
    X, Y, pos = get_meshgrid_and_pos(xlims, ylims, ngrid)
    fXY = rvXY.pdf(pos)

    # Figure grid
    gs = gridspec.GridSpec(2, 2, width_ratios=[6,1], height_ratios=[1,4])

    # Contour plot of f_XY
    ax = plt.subplot(gs[1,0])    
    cax = ax.contourf(fXY, origin = 'lower',
                      extent=(*xlims, *ylims),
                      levels=12,
                      cmap="Greys")
    ax.set_xlabel('$x$')
    ax.set_ylabel('$y$')
    ax.text(5, 6, "$f_{XY}$", fontsize="x-large")

    # Compute marginal distributions
    fYm, fXm = margins(fXY)
    dx = (xlims[1] - xlims[0]) / (ngrid - 1)
    dy = (ylims[1] - ylims[0]) / (ngrid - 1)
    fX = fXm.flatten() * dy
    fY = fYm.flatten() * dx

    # The marginal f_X (top)
    xs = X[0]
    axt = plt.subplot(gs[0,0], sharex=ax, frameon=False, xlim=xlims, ylim=(0, 1.1*fX.max()))
    axt.plot(xs, fX, color = 'black')
    axt.fill_between(xs, 0, fX, alpha=.5, color = 'gray')
    axt.tick_params(labelbottom=False)
    axt.tick_params(labelleft=False)
    axt.text(5, 0.08, "$f_{X}$", fontsize="x-large")

    # The marginal f_Y (right)
    ys = Y[:,0]
    axr = plt.subplot(gs[1,1], sharey=ax, frameon=False, xlim=(0, 1.05*fY.max()), ylim=ylims)
    axr.plot(0*np.ones_like(ys), ys)
    axr.plot(fY, ys, color = 'black')
    axr.fill_betweenx(ys, 0, fY, alpha=0.5, color="gray")
    axr.tick_params(labelbottom=False)
    axr.tick_params(labelleft=False)
    axr.text(0.3,3.2, "$f_{Y}$", fontsize="x-large")

    return fig


def find_nearest1(array, value):
    """
    Find the index of the `array` entry that is closest to `value`.
    Helper function used by `plot_slices_through_joint_pdf` et al.
    """
    idx, _ = min(enumerate(array), key=lambda x: abs(x[1]-value))
    return idx


def polygon_under_graph(xs, ys):
    """
    Construct the vertex list that defines the polygon filling the space
    under the curve that passes through the points `[(x,y) in zip(xs,ys)]`.
    Helper function used by `plot_slices_through_joint_pdf` et al.
    """
    return [(xs[0], 0.), *zip(xs, ys), (xs[-1], 0.)]


def plot_slices_through_joint_pdf(rvXY, xlims, ylims, xcuts, ngrid=500, fig=None):
    """
    Plot slices through the joint distribution $f_XY$ at the x-values in `xcuts`.
    """
    # Setup figure and axes
    if fig is None:
        fig = plt.figure(figsize=(7,4))
    ax = fig.add_subplot(projection='3d')

    # Compute the joint-probability density function values
    X, Y, pos = get_meshgrid_and_pos(xlims, ylims, ngrid)
    xs, ys = X[0], Y[:,0]
    fXY = rvXY.pdf(pos)

    # The entry `verts[i]` is a list of (x,y) pairs defining polygon `i`
    verts = []
    for xcut in xcuts:
        xidx = find_nearest1(xs, xcut)
        fXY_at_xcut = fXY[xidx,:]
        vert = polygon_under_graph(ys, fXY_at_xcut)
        verts.append(vert)
    
    # Plot polygons
    facecolors = plt.colormaps['viridis_r'](np.linspace(0, 1, len(verts)))
    poly = PolyCollection(verts, facecolors=facecolors, alpha=.6)
    ax.add_collection3d(poly, zs=xcuts, zdir='x')
    ax.set_box_aspect((9, 5, 4))
    zmax = 0.06
    ax.set(xlim=xlims, ylim=ylims, zlim=(0, zmax), xlabel='$x$', ylabel='$y$', zlabel='probability')
    ax.set_xticks(range(4,17,1))

    return fig


def plot_conditional_fYgivenX(rvXY, xlims, ylims, xcuts, ngrid=500, fig=None):
    """
    Plot the conditional distribution $f_Y|X$ at the x-values in `xcuts`.
    """
    # Setup figure and axes
    if fig is None:
        fig = plt.figure(figsize=(7,4))
    ax = fig.add_subplot(projection='3d')

    # Compute the joint-probability density function values
    X, Y, pos = get_meshgrid_and_pos(xlims, ylims, ngrid)
    xs, ys = X[0], Y[:,0]
    fXY = rvXY.pdf(pos)

    # hack to find normalizing height
    xmid = xcuts[len(xcuts)//2]
    xmididx = find_nearest1(xs, xmid)
    fXY_at_xmid = fXY[xmididx,:]
    fYgiven_xmid = fXY_at_xmid / np.sum(fXY_at_xmid)
    maxfYgiven_xmid = max(fYgiven_xmid)

    # The entry `verts[i]` is a list of (x,y) pairs defining polygon `i`
    verts = []
    for xcut in xcuts:
        xidx = find_nearest1(xs, xcut)
        fXY_at_xcut = fXY[xidx,:]
        fYgiven_xcut = fXY_at_xcut / np.sum(fXY_at_xcut)
        # hack to normalize height
        zscale = max(fYgiven_xcut) / maxfYgiven_xmid
        fYgiven_xcut = fYgiven_xcut / zscale
        # /hack to normalize height
        vert = polygon_under_graph(ys, fYgiven_xcut)
        verts.append(vert)

    # Plot polygons
    facecolors = plt.colormaps['viridis_r'](np.linspace(0, 1, len(verts)))
    poly = PolyCollection(verts, facecolors=facecolors, alpha=.7)
    ax.add_collection3d(poly, zs=xcuts, zdir='x')
    ax.set_box_aspect((9, 5, 4))
    zmax = 0.006
    ax.set(xlim=xlims, ylim=ylims, zlim=(0, zmax), xlabel='$x$', ylabel='$y$', zlabel='probability');
    ax.set_xticks(range(4, 17, 1))

    return fig




# Section 2.4: Bulk and tails of a continuous distribution
################################################################################


def bulk_of_pdf_panel(rvX, rv_name, xlims, xticks=None, ns=[1,2,3], fig=None):
    """
    Print a 1x3 panel figure highlighting the probability mass that lies within
    `ns` standard deviations from the mean of the random variable `rvX`.
    """
    if fig is None:
        fig, axs = plt.subplots(1, 3, figsize=(9.2,2), sharey=True)
    else:
        axs = fig.subplots(1, 3, sharey=True)

    muX = rvX.mean()    # mean of the random variable rvX
    sigmaX = rvX.std()  # standard deviation of rvX

    for i, n in enumerate(ns):
        ax = axs[i]
        bulk_interval = [muX - n*sigmaX, muX + n*sigmaX]
        letter = ["a", "b", "c"][i]
        mu = "\\mu_" + rv_name
        sigma = "\\sigma_" + rv_name
        if n == 1:
            title = f"({letter}) Pr($\\{{{mu}-{sigma} \\leq {rv_name} \\leq {mu}+{sigma}\\}}$)"
        else:
            title = f"({letter}) Pr($\\{{{mu}-{n}{sigma} \\leq {rv_name} \\leq {mu}+{n}{sigma}\\}}$)"
        calc_prob_and_plot(rvX, *bulk_interval, xlims=xlims, ax=ax, title=title)
        if xticks:
            ax.set_xticks(xticks)

    return fig



def tails_of_pdf_panel(rvX, rv_name, xlims, xticks=None, ns=[1,2,3], fig=None):
    """
    Print a 1x3 panel figure highlighting the probability mass that lies within
    `ns` standard deviations from the mean of the random variable `rvX`.
    """
    if fig is None:
        fig, axs = plt.subplots(1, 3, figsize=(9.2,2), sharey=True)
    else:
        axs = fig.subplots(1, 3, sharey=True)

    muX = rvX.mean()    # mean of the random variable rvX
    sigmaX = rvX.std()  # standard deviation of rvX

    for i, n in enumerate(ns):
        ax = axs[i]
        x_l = muX - n * sigmaX
        x_r = muX + n * sigmaX
        letter = ["a", "b", "c"][i]
        mu = "\\mu_" + rv_name
        sigma = "\\sigma_" + rv_name
        if n == 1:
            title = f"({letter}) Pr($\\{{{rv_name} \\leq {mu}-{sigma}\\}} \\cup \\{{{rv_name} \\geq {mu}+{sigma}\\}}$)"
        else:
            title = f"({letter}) Pr($\\{{{rv_name} \\leq {mu}-{n}{sigma}\\}} \\cup \\{{{rv_name} \\geq {mu}+{n}{sigma}\\}}$)"
        calc_prob_and_plot_tails(rvX, x_l, x_r, xlims=xlims, ax=ax, title=title)
        if xticks:
            ax.set_xticks(xticks)

    return fig














# Section 3.4: Tails of a sampling distributions
################################################################################

def plot_panel_pvalue_theta0_tails(figsize=(9,2.3)):
    rvT = tdist(df=9)
    xs = np.linspace(-4, 4, 1000)
    ys = rvT.pdf(xs)

    with plt.rc_context({"figure.figsize":figsize}), sns.axes_style("ticks"):
        fig, (ax3, ax1, ax2) = plt.subplots(1,3)
        ax3.set_ylabel(r"$f_{\widehat{\Theta}_0}$")

        # RIGHT
        title = '(a) right-tailed test'
        ax3.set_title(title, fontsize=13)#, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax3)
        ax3.set_xlim(-4, 4)
        ax3.set_ylim(0, 0.42)
        ax3.set_xticks([0,2])
        ax3.set_xticklabels([])
        ax3.set_yticks([])
        ax3.spines[['right', 'top']].set_visible(False)

        # highlight the right tail
        mask = (xs > 2)
        ax3.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax3.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=0.5, color="red")
        ax3.text(2, -0.03, r"$\hat{\theta}_{\mathbf{x}}$", va="top", ha="center")
        ax3.text(0, -0.04, r"$\theta_0$", va="top", ha="center")


        # LEFT
        title = '(b) left-tailed test'
        ax1.set_title(title, fontsize=13) #, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax1)
        ax1.set_xlim(-4, 4)
        ax1.set_ylim(0, 0.42)
        ax1.set_xticks([-2,0])
        ax1.set_xticklabels([])
        ax1.set_yticks([])
        ax1.spines[['left', 'right', 'top']].set_visible(False)

        # highlight the left tail
        mask = (xs < -2)
        ax1.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax1.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=0.5, color="red")
        ax1.text(-2, -0.03, r"$\hat{\theta}_{\mathbf{x}}$", va="top", ha="center")
        ax1.text(0, -0.04, r"$\theta_0$", va="top", ha="center")

        # TWO-TAILED
        title = '(c) two-tailed test'
        ax2.set_title(title, fontsize=13)#, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax2)
        ax2.set_xlim(-4, 4)
        ax2.set_ylim(0, 0.42)
        ax2.set_xticks([-2,0,2])
        ax2.set_xticklabels([])
        ax2.set_yticks([])
        ax2.spines[['left', 'right', 'top']].set_visible(False)

        # highlight the left and right tails
        mask = (xs < -2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=0.5, color="red")
        # ax2.text(-2, -0.03, r"$-|\hat{\theta}_{\mathbf{x}}|$", va="top", ha="center")
        ax2.text(-2, -0.03, r"$\theta_0$-dev", va="top", ha="center")
        mask = (xs > 2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=0.5, color="red")
        ax2.text(2, -0.03, r"$\theta_0$+dev", va="top", ha="center")
        ax2.text(0, -0.04, r"$\theta_0$", va="top", ha="center")

    return fig



def plot_panel_pvalue_t_tails(figsize=(9,2.3)):    
    rvT = tdist(df=9)
    xs = np.linspace(-4, 4, 1000)
    ys = rvT.pdf(xs)

    with plt.rc_context({"figure.figsize":figsize}), sns.axes_style("ticks"):
        fig, (ax3, ax1, ax2) = plt.subplots(1,3)
        ax3.set_ylabel("$f_{T_0}$")

        # RIGHT
        title = '(a) right-tailed test'
        ax3.set_title(title, fontsize=13)#, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax3)
        ax3.set_xlim(-4, 4)
        ax3.set_ylim(0, 0.42)
        ax3.set_xticks([0,2])
        ax3.set_xticklabels([])
        ax3.set_yticks([])
        ax3.spines[['right', 'top']].set_visible(False)

        # highlight the right tail
        mask = (xs > 2)
        ax3.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax3.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=0.5, color="red")
        ax3.text(2, -0.03, r"$t_{\mathbf{x}}$", va="top", ha="center")
        ax3.text(0, -0.04, r"$0$", va="top", ha="center")

        # LEFT
        title = '(b) left-tailed test'
        ax1.set_title(title, fontsize=13) #, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax1)
        ax1.set_xlim(-4, 4)
        ax1.set_ylim(0, 0.42)
        ax1.set_xticks([-2,0])
        ax1.set_xticklabels([])
        ax1.set_yticks([])
        ax1.spines[['left', 'right', 'top']].set_visible(False)

        # highlight the left tail
        mask = (xs < -2)
        ax1.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax1.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=0.5, color="red")
        ax1.text(-2, -0.03, r"$t_{\mathbf{x}}$", va="top", ha="center")
        ax1.text(0, -0.04, r"$0$", va="top", ha="center")

        # TWO-TAILED
        title = '(c) two-tailed test'
        ax2.set_title(title, fontsize=13)#, y=-0.26)
        sns.lineplot(x=xs, y=ys, ax=ax2)
        ax2.set_xlim(-4, 4)
        ax2.set_ylim(0, 0.42)
        ax2.set_xticks([-2,0,2])
        ax2.set_xticklabels([])
        ax2.set_yticks([])
        ax2.spines[['left', 'right', 'top']].set_visible(False)

        # highlight the left and right tails
        mask = (xs < -2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=0.5, color="red")
        ax2.text(-2, -0.03, r"$-|t_{\mathbf{x}}|$", va="top", ha="center")
        mask = (xs > 2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=0.5, color="red")
        ax2.text(2, -0.03, r"$|t_{\mathbf{x}}|$", va="top", ha="center")
        ax2.text(0, -0.04, r"$0$", va="top", ha="center")

    return fig





# Section 3.5: Tails of a sampling distributions
################################################################################


def plot_panel_pvalue_D0_hist_and_dist(figsize=(7,2)):
    rvT = tdist(df=9)
    xs = np.linspace(-4, 4, 1000)
    ys = rvT.pdf(xs)
    N = 100000
    np.random.seed(42)
    ts = rvT.rvs(N)
    bins = nicebins(xs, 2, nbins=50)
    with plt.rc_context({"figure.figsize":figsize}), sns.axes_style("ticks"):
        fig, (ax1, ax2) = plt.subplots(1,2)

        # D0 hist
        sns.histplot(ts, ax=ax1, bins=bins, alpha=0.3)
        ax1.set_title("(a) permutation test")
        ax1.set_xlim(-4, 4)
        ax1.set_xticks([2])
        ax1.set_xticklabels([])
        ax1.set_yticks([])
        ax1.set_ylabel(r"$f_{\widehat{D}_0}$")
        # highlight the left and right tails
        tailvaluesl = [t for t in ts if t <= -2]
        sns.histplot(tailvaluesl, bins=bins, ax=ax1, color="red")
        ax1.text(-2.3, -630, r"$-|\hat{d}|$", verticalalignment="top", horizontalalignment="center")
        tailvaluesr = [t for t in ts if t >= 2]
        sns.histplot(tailvaluesr, bins=bins, ax=ax1, color="red")
        ax1.text(2, -630, r"$|\hat{d}|$", verticalalignment="top", horizontalalignment="center")
        # ax1.axvline(2, color="red")
        
        # D0 dist
        sns.lineplot(x=xs, y=ys, ax=ax2)
        ax2.set_title("(b) analytical apprixmation")
        ax2.set_xlim(-4, 4)
        ax2.set_ylim(0, 0.42)
        ax2.set_xticks([-2,2])
        ax2.set_xticklabels([])
        ax2.set_yticks([])
        ax2.set_ylabel(r"$f_{\widehat{D}_0}$")
        # highlight the left and right tails
        mask = (xs < -2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.text(-2, -0.03, r"$-|\hat{d}|$", verticalalignment="top", horizontalalignment="center")
        mask = (xs > 2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.text(2, -0.03, r"$|\hat{d}|$", verticalalignment="top", horizontalalignment="center")

    return fig



def plot_panel_pvalue_D0_and_T0(figsize=(7,2)):
    from scipy.stats import t as tdist
    rvT = tdist(df=9)
    xs = np.linspace(-4, 4, 1000)
    ys = rvT.pdf(xs)
    with plt.rc_context({"figure.figsize":figsize}), sns.axes_style("ticks"):
        fig, (ax1, ax2) = plt.subplots(1,2)

        # D0
        sns.lineplot(x=xs, y=ys, ax=ax1)
        ax1.set_xlim(-4, 4)
        ax1.set_ylim(0, 0.42)
        ax1.set_xticks([2])
        ax1.set_xticklabels([])
        ax1.set_yticks([])
        ax1.set_ylabel(r"$f_{\widehat{D}_0}$")
        # highlight the left and right tails
        mask = (xs < -2)
        ax1.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax1.text(-2, -0.03, r"$-|\hat{d}|$", verticalalignment="top", horizontalalignment="center")
        mask = (xs > 2)
        ax1.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax1.text(2, -0.03, r"$|\hat{d}|$", verticalalignment="top", horizontalalignment="center")

        # T0
        sns.lineplot(x=xs, y=ys, ax=ax2)
        ax2.set_xlim(-4, 4)
        ax2.set_ylim(0, 0.42)
        ax2.set_xticks([-2,2])
        ax2.set_xticklabels([])
        ax2.set_yticks([])
        ax2.set_ylabel("$f_{T_0}$")
        # highlight the left and right tails
        mask = (xs < -2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.text(-2, -0.03, r"$-|t|$", verticalalignment="top", horizontalalignment="center")
        mask = (xs > 2)
        ax2.fill_between(xs[mask], y1=ys[mask], alpha=0.6, facecolor="red")
        ax2.text(2, -0.03, r"$|t|$", verticalalignment="top", horizontalalignment="center")

    return fig


# Section 3.6 — Statistical design and error analysis
################################################################################



def plot_rejection_region(ax, xs, ys, rvT, alt, title):
    # design choices
    transp = 0.3
    alpha_color = "#4A25FF"
    beta_color = "#0CB0D6"
    axis_color = "#808080"

    ax.set_title(title, fontsize=11)

    # manually add arrowhead to x-axis + label t at the end
    ax.plot(1, 0, ">", color=axis_color, transform=ax.get_yaxis_transform(), clip_on=False)
    ax.set_xlabel("t")
    ax.xaxis.set_label_coords(1, 0.2)

    sns.lineplot(x=xs, y=ys, ax=ax, color="k")
    ax.set_xlim(-4, 4)
    ax.set_ylim(0, 0.42)
    ax.set_xticklabels([])
    ax.set_yticks([])
    ax.spines[['left', 'right', 'top']].set_visible(False)
    ax.spines['bottom'].set_color(axis_color)
    ax.tick_params(axis='x', colors=axis_color)
    ax.xaxis.label.set_color(axis_color)
    
    if alt == "greater":
        ax.set_xticks([2])
        # highlight the right tail
        mask = (xs > 2)
        ax.fill_between(xs[mask], y1=ys[mask], alpha=transp, facecolor=alpha_color)
        ax.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=transp+0.2, color=alpha_color)
        ax.text(2, -0.03, r"$\mathrm{CV}_{\alpha}^+$", verticalalignment="top", horizontalalignment="center")

    elif alt == "less":
        ax.set_xticks([-2])
        # highlight the left tail
        mask = (xs < -2)
        ax.fill_between(xs[mask], y1=ys[mask], alpha=transp, facecolor=alpha_color)
        ax.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=transp+0.2, color=alpha_color)
        ax.text(-2, -0.03, r"$\mathrm{CV}_{\alpha}^-$", verticalalignment="top", horizontalalignment="center")

    elif alt == "two-sided":
        ax.set_xticks([-2,2])
        # highlight the left and right tails
        mask = (xs < -2)
        ax.fill_between(xs[mask], y1=ys[mask], alpha=transp, facecolor=alpha_color)
        ax.vlines([-2], ymin=0, ymax=rvT.pdf(-2), linestyle="-", alpha=transp+0.2, color=alpha_color)
        ax.text(-2, -0.03, r"$\mathrm{CV}_{\alpha/2}^-$", verticalalignment="top", horizontalalignment="center")
        mask = (xs > 2)
        ax.fill_between(xs[mask], y1=ys[mask], alpha=transp, facecolor=alpha_color)
        ax.vlines([2], ymin=0, ymax=rvT.pdf(2), linestyle="-", alpha=transp+0.2, color=alpha_color)
        ax.text(2, -0.03, r"$\mathrm{CV}_{\alpha/2}^+$", verticalalignment="top", horizontalalignment="center")



def plot_panel_rejection_regions(figsize=(7,1.6)):
    rvT = tdist(df=9)
    xs = np.linspace(-4, 4, 1000)
    ys = rvT.pdf(xs)
    with sns.axes_style("ticks"), plt.rc_context({"figure.figsize":figsize}):
        fig, (ax3, ax1, ax2) = plt.subplots(1,3)
        # RIGHT
        title = '(a) right-tailed rejection region'
        plot_rejection_region(ax3, xs, ys, rvT, "greater", title)
        # LEFT
        title = '(b) left-tailed rejection region'
        plot_rejection_region(ax1, xs, ys, rvT, "less", title)
        # TWO-TAILED
        title = '(c) two-tailed rejection region'
        plot_rejection_region(ax2, xs, ys, rvT, "two-sided", title)
        fig.tight_layout()
        return fig


# Hierarchical models
################################################################################

def plot_counties(radon, idata_cp=None, idata_np=None, idata_pp=None, idata_pp2=None,
                  figsize=None, counties=None):
    """
    Generate a 2x4 panel of scatter plots for the `selected_counties`
    and optional line plots models:
    - `idata_cp`: complete pooling model
    - `idata_np`: no pooling model
    - `idata_pp`: partial pooling model (varying intercepts)
    - `idata_pp2`: partial pooling model (varying slopes and intercepts)
    """
    if counties == None:
        counties = [
            "LAC QUI PARLE",
            "AITKIN",
            "KOOCHICHING",
            "DOUGLAS",
            "HENNEPIN",
            "STEARNS",
            "RAMSEY",
            "ST LOUIS",
        ]

    if idata_cp:
        # completely pooled means
        post1_means = idata_cp["posterior"].mean(dim=("chain", "draw"))

    if idata_np:
        # no pooling means
        post2_means = idata_np["posterior"].mean(dim=("chain", "draw"))

    if idata_pp:
        # partial pooling model (varying intercepts)
        post3_means = idata_pp["posterior"].mean(dim=("chain", "draw"))
    
    if idata_pp2:
        # partial pooling model (varying slopes and intercepts)
        post4_means = idata_pp2["posterior"].mean(dim=("chain", "draw"))

    n_rows = math.ceil(len(counties) / 4)
    if figsize is None:
        if n_rows == 1:
            figsize = (10,2)
        elif n_rows == 2:
            figsize = (10,4)
        if n_rows > 2:
            figsize = (10, 2*n_rows)
    fig, axes = plt.subplots(n_rows, 4, figsize=figsize, sharey=True, sharex=True)
    axes = axes.flatten()
    
    for i, c in enumerate(counties):
        y = radon.log_radon[radon.county == c]
        x = radon.floor[radon.county == c]
        x = x.map({"basement":0, "ground":1})
        axes[i].scatter(x + np.random.randn(len(x)) * 0.01, y, alpha=0.4)

        # linspace of x-values 
        xvals = xr.DataArray(np.linspace(0, 1))

        if idata_cp:
            # Plot complete pooling model
            model1_vals = post1_means["Intercept"] + post1_means["floor"].values*xvals
            axes[i].plot(xvals, model1_vals, "C0-")

        if idata_np: 
            # Plot no pooling model
            b = post2_means["county"].sel(county_dim=c)
            m = post2_means["floor"]
            axes[i].plot(xvals, b.values + m.values*xvals, "C1--")

        if idata_pp:
            # Plot varying intercepts model
            post3c = post3_means.sel(county__factor_dim=c)
            # When using 0 + floor model
            # slope = post.floor.values[1] - post.floor.values[0]
            # theta = post["1|county"].values + post.floor.values[0] + slope * xvals
            # When using 1 + floor model
            slope = post3c["floor"].values[0]
            theta = post3c["Intercept"] + post3c["1|county"].values + slope*xvals
            axes[i].plot(xvals, theta, "k:")

        if idata_pp2:
            # Plot varying slopes and intercepts model
            post4c = post4_means.sel(county__factor_dim=c)
            intercept = post4c["Intercept"] + post4c["1|county"].values
            slope = post4c["floor"].values[0] + post4c["floor|county"].values[0]
            theta = intercept + slope*xvals
            axes[i].plot(xvals, theta, "C3-.")

        axes[i].set_xticks([0, 1])
        axes[i].set_xticklabels(["basement", "ground"])
        axes[i].set_ylim(-1, 3)
        axes[i].set_title(c)
        if i % 4 == 0:
            axes[i].set_ylabel("log radon level")

    return fig


from collections import defaultdict
import os
import time

import arviz as az
from arviz.plots.plot_utils import calculate_point_estimate as calc_point_est
import bambi as bmb
import numpy as np
import pandas as pd
import pingouin as pg
from scipy.stats import norm
from scipy.stats import ttest_ind

from ..bayes import calc_dmeans_stats
from ..bayes import hdi_from_idata
from ..confidence_intervals import ci_dmeans
from ..hypothesis_tests import permutation_test_dmeans
from ..hypothesis_tests import ttest_dmeans
from ..utils import loglevel


# BAYESIAN DMEANS SENSITIVITY ANALYSIS
################################################################################

def fit_bayesian_model_iqs2(iqs2, new_priors, random_seed=42):
    """
    Fits the model like Example 2 in Section 5.4 with `new_priors`
    overwriting the defaults priors in `priors2`.
    Returns the `idata2` object.
    """
    priors2 = {
        "group": bmb.Prior("Normal", mu=100, sigma=35),
        "sigma": {
            "group": bmb.Prior("Normal", mu=1, sigma=2)
        },
        "nu": bmb.Prior("Gamma", alpha=2, beta=0.1),
    }
    priors2.update(new_priors)
    formula2 = bmb.Formula("iq ~ 0 + group", "sigma ~ 0 + group")
    mod2 = bmb.Model(formula=formula2,
                     family="t",
                     link={"mu": "identity", "sigma": "log"},
                     priors=priors2,
                     data=iqs2)
    with loglevel("ERROR", module="pymc"):
        idata2 = mod2.fit(draws=2000, random_seed=random_seed, progressbar=False)
    return idata2



def sens_analysis_dmeans_iqs2(iqs2):
    """
    Generate the table showing the results of the sensitivity analysis
    for Example 2 in Section 5.4.
    """
    # Define the priors
    M_priors = {
         "orig": {
              "display": r"$\mathcal{N}(100,35)$",
              "bambi": bmb.Prior("Normal", mu=100, sigma=35),
         },
         "wider": {
              "display": r"$\mathcal{N}(100,50)$",
              "bambi": bmb.Prior("Normal", mu=100, sigma=50),
         },
         "tighter": {
              "display": r"$\mathcal{N}(100,10)$",
              "bambi": bmb.Prior("Normal", mu=100, sigma=10),
         },
    }
    logSigma_priors = {
        "orig": {
            "display": r"$\mathcal{N}(1,2)$",
            "bambi": bmb.Prior("Normal", mu=1, sigma=2),
        },
        "low": {
            "display": r"$\mathcal{N}(0,1)$",
            "bambi": bmb.Prior("Normal", mu=0, sigma=1),
        },
    }
    Nu_priors = {
        "orig": {
            "display": r"$\Gamma(2,0.1)$",
            "bambi": bmb.Prior("Gamma", alpha=2, beta=0.1),
        },
        "expon": {
            "display": r"$\textrm{Expon}(1/30)$",
            "bambi": bmb.Prior("Exponential", lam=1/30),
        },
    }
    
    experiments = [
        dict(name="orig", mean="orig",    sigma="orig", nu="orig",  seed=42),
        dict(name="orig", mean="wider",   sigma="orig", nu="orig",  seed=43),
        dict(name="orig", mean="tighter", sigma="orig", nu="orig",  seed=44),
        dict(name="orig", mean="orig",    sigma="low",  nu="orig",  seed=45),
        dict(name="orig", mean="orig",    sigma="orig", nu="expon", seed=46),
    ]

    # Prepare results table
    experiemnt_columns = [
        "M_prior",
        "logSigma_prior",
        "Nu_prior",
    ]
    result_columns = [
        "dmeans_mean",
        "dmeans_95hdi",
        "dsigmas_mode",
        "dsigmas_95hdi",
        "nu_mode",
        "codhend_mode",
    ]
    results_columns = experiemnt_columns + result_columns
    results_rows = range(len(experiments))
    results = pd.DataFrame(index=results_rows, columns=results_columns)

    for i, exp in enumerate(experiments):
        # print("fitting model", i, "...")
        priors = {}  # priors to be used for current run

        # Set priors based on specification in `exp`
        results.loc[i, "M_prior"] = M_priors[exp["mean"]]["display"]
        priors["group"] = M_priors[exp["mean"]]["bambi"]
        results.loc[i, "logSigma_prior"] = logSigma_priors[exp["sigma"]]["display"]
        priors["sigma"] = {"group": logSigma_priors[exp["sigma"]]["bambi"]}
        results.loc[i, "Nu_prior"] = Nu_priors[exp["nu"]]["display"]
        priors["nu"] = Nu_priors[exp["nu"]]["bambi"]

        # Fit model
        idata2 = fit_bayesian_model_iqs2(iqs2, priors, random_seed=exp["seed"])
        calc_dmeans_stats(idata2, group_name="group")

        # Calculate results
        post2 = idata2["posterior"]
        summary2 = az.summary(post2, kind="stats", hdi_prob=0.95)
        ## Calculate dmeans_mean
        results.loc[i, "dmeans_mean"] = summary2.loc["dmeans", "mean"]
        ## Calculate dmeans_95hdi
        dmeans_ci_low = summary2.loc["dmeans", "hdi_2.5%"]
        dmeans_ci_high = summary2.loc["dmeans","hdi_97.5%"]
        results.loc[i, "dmeans_95hdi"] = [dmeans_ci_low, dmeans_ci_high]
        ## Calculate dsigmas_mode
        dsigmas = post2["dsigmas"].values.flatten()
        results.loc[i, "dsigmas_mode"] = calc_point_est("mode", dsigmas).round(3)
        ## Calculate dsigmas_95hdi
        dsigmas_ci_low = summary2.loc["dsigmas", "hdi_2.5%"]
        dsigmas_ci_high = summary2.loc["dsigmas","hdi_97.5%"]
        results.loc[i, "dsigmas_95hdi"] = [dsigmas_ci_low, dsigmas_ci_high]
        ## Calculate nu_mode
        nus = post2["nu"].values.flatten()
        results.loc[i, "nu_mode"] = calc_point_est("mode", nus).round(3)
        ## Calculate codhend_mode
        cohends = post2["dsigmas"].values.flatten()
        results.loc[i, "codhend_mode"] = calc_point_est("mode", cohends).round(3)

    return results






# DMEANS PERFORMANCE ANALYSIS
################################################################################

def gen_dmeans_dataset(n, Delta, prop_outliers=0, random_seed=42):
    """
    Generate a dataset with two-groups of size `n`.
    The mean of the control group is `0`,
    the mean of the treatment group is `Delta`.
    We'll make a proportion `prop_outliers` of the values
    in each group outliers, coming from a population
    with much wider standard deviation.
    """
    np.random.seed(random_seed)
    
    # Generate control group
    controls = norm(0, 1).rvs(n)

    # Generate treated group
    treated = norm(0 + Delta, 1).rvs(n)

    # Add outliers
    n_outliers = int(prop_outliers * n)
    control_outliers = norm(0, 5).rvs(n_outliers)  
    controls[0:n_outliers] = control_outliers
    treated_outliers = norm(0 + Delta, 5).rvs(n_outliers) 
    treated[0:n_outliers] = treated_outliers

    # Package dataset as a Pandas DataFrame
    groups = ["treat"]*len(treated) + ["ctrl"]*len(controls)
    values = np.concatenate((treated, controls))
    dataset = pd.DataFrame({"group": groups, "value": values})
    return dataset


def gen_dmeans_datasets(ns, Deltas, outliers_options, random_seed_start=45):
    """
	This function prepares a dictionary of datasets that exhibit different
	combinations of the characteristics we might encounter in the real world.
	To keep things simple,
	we'll assume both populations have unit standard deviation sigma_A = sigma_B = 1
	and group A has mean mu_A=0.
    - Sample size: n=20, n=30, n=50, or n=100
    - Effect size Delta: 0, 0.2, 0.5, 1.0,
	  which means Group B has mean mu_B= 0, 0.2, 0.5, 0.8, 1.3.
    - Outliers: no outliers, some outliers, lots of outliers.
    We'll generate samples from each combination of conditions,
    then use various models to perform the hypothesis tests and count
    the number of times we reach the correct decision.
    """
    random_seed = random_seed_start
    dataset_specs = []
    for n in ns:
        for Delta in Deltas:
            for outliers in outliers_options:
                spec = dict(n=n,  Delta=Delta, outliers=outliers, random_seed=random_seed)
                dataset_specs.append(spec)
                random_seed += 1

    return dataset_specs



def is_inside(val, interval):
    """
    Check if the value `val` is inside the interval `interval`.
    """
    if val >= interval[0] and val <= interval[1]:
        return True
    else:
        return False


def calc_dmeans(idata, group_name="group", groups=["ctrl", "treat"]):
    """
    Simplified version of `ministats.bayes.calc_dmeans_stats` used to
    calculate `dmeans` posterior from difference of group means.
    """
    group_dim = group_name + "_dim"
    post = idata["posterior"]
    post["mu_" + groups[0]] = post[group_name].loc[{group_dim:groups[0]}]
    post["mu_" + groups[1]] = post[group_name].loc[{group_dim:groups[1]}]
    post["dmeans"] = post["mu_" + groups[1]] - post["mu_" + groups[0]]



def fit_dmeans_models(dataset, random_seed=42):
    """
    Fit the following models for the difference of two means:
    - Permutation test from Section 3.5
    - Welch's two-sample $t$-test from Section 3.5
    - Bayesian model that uses normal as data model
    - Robust Bayesian model that uses t-distribution as data model
    - Bayes factor using JZS prior (using `pingouin` library)
    For each model, we run the hypothesis test to decide if the populations
    are the same or different based on the conventional cutoff level of 5%.
    We also construct a 90% interval estimates for the unknown `Delta`.
    """
    models = ["perm", "welch", "norm_bayes", "robust_bayes", "bf"]

    # Helper functions to get decision in each situation
    def get_pval_decision(pval, alpha=0.05):
        if pval <= alpha:
            return "reject H0"
        else:
            return "fail to reject H0"

    def get_ci_decision(val, interval):
        if is_inside(val, interval):
            return "fail to reject H0"
        else:
            return "reject H0"

    def get_bf_decision(bfA0, cutoff_reject_H0=3, cutoff_accept_H0=1/3):
        if bfA0 >= cutoff_reject_H0:
            return "reject H0"
        elif bfA0 <= cutoff_accept_H0:
            return "accept H0"
        else:
            return "no decision"

    treated = dataset[dataset["group"]=="treat"]["value"].values
    controls = dataset[dataset["group"]=="ctrl"]["value"].values

    # Fit the models
    results = {}
    if "perm" in models:
        np.random.seed(random_seed)
        pval = permutation_test_dmeans(treated, controls)
        decision = get_pval_decision(pval)
        ci90 = ci_dmeans(treated, controls, alpha=0.1, method="b")
        results["perm"] = {"decision": decision, "ci90": ci90}

    if "welch" in models:
        pval = ttest_dmeans(treated, controls, equal_var=False, alt="two-sided")
        decision = get_pval_decision(pval)
        ci90 = ci_dmeans(treated, controls, alpha=0.1, method="a")
        results["welch"] = {"decision": decision, "ci90": ci90}

    if "norm_bayes" in models:
        priors = {
            "group": bmb.Prior("Normal", mu=0, sigma=2),
            "sigma": {
                "group": bmb.Prior("LogNormal", mu=0, sigma=1)
                # "group": bmb.Prior("Normal", mu=1, sigma=1)
            }
        }
        formula = bmb.Formula("value ~ 0 + group", "sigma ~ 0 + group")
        norm_mod = bmb.Model(formula=formula, family="gaussian", priors=priors, data=dataset)
        with loglevel("ERROR", module="pymc"):
            idata = norm_mod.fit(draws=2000, random_seed=random_seed, progressbar=False)
        calc_dmeans(idata)
        ci95 = hdi_from_idata(idata, var_name="dmeans", hdi_prob=0.95)
        decision = get_ci_decision(0, ci95)
        ci90 = hdi_from_idata(idata, var_name="dmeans", hdi_prob=0.9)
        results["norm_bayes"] = {"decision": decision, "ci90": ci90}

    if "robust_bayes" in models:
        priors = {
            "group": bmb.Prior("Normal", mu=0, sigma=2),
            "sigma": {
                "group": bmb.Prior("Normal", mu=0, sigma=1)
            },
            "nu": bmb.Prior("Gamma", alpha=2, beta=0.1),
        }
        formula = bmb.Formula("value ~ 0 + group", "sigma ~ 0 + group")
        robust_mod = bmb.Model(formula=formula, family="t", priors=priors, data=dataset)
        with loglevel("ERROR", module="pymc"):
            idata = robust_mod.fit(draws=2000, random_seed=random_seed, progressbar=False)
        calc_dmeans(idata)
        ci95 = hdi_from_idata(idata, var_name="dmeans", hdi_prob=0.95)
        decision = get_ci_decision(0, ci95)
        ci90 = hdi_from_idata(idata, var_name="dmeans", hdi_prob=0.9)
        results["robust_bayes"] = {"decision": decision, "ci90": ci90}

    if "bf" in models:
        ttres = ttest_ind(treated, controls, equal_var=True)
        n, m = len(treated), len(controls)
        bfA0 = pg.bayesfactor_ttest(ttres.statistic, nx=n, ny=m, r=0.707)
        decision = get_bf_decision(bfA0)
        ci90 = None
        results["bf"] = {"decision": decision, "ci90": ci90}

    return results



def calc_dmeans_perf_metrics(
        ns=[20,30,50,100],
        Deltas=[0, 0.2, 0.5, 0.8, 1.3],
        outliers_options=["no", "few", "lots"],
        reps=100,
        random_seed_start=45):
    """
    Calculate the performance of the different statistical analysis procedures
    for each dataset specification, based on `reps` repeated samples.
    We'll count the following outcomes:    
    - `count_reject`: number of times H_0: Delta=0 is rejected
    - `count_fail_to_reject`: number of times we fail to reject H_0: Delta=0
    - `count_captured`: when the interval estimate (ci or hdi) contain the true `Delta`
    - `avg_width`: average width of the interval estimate (ci or hdi) over all `reps`
    """
    dataset_specs = gen_dmeans_datasets(ns=ns,
                                        Deltas=Deltas,
                                        outliers_options=outliers_options,
                                        random_seed_start=random_seed_start)
    print("Simulating a total of", len(dataset_specs), "dataset specs")

    dataset_columns = [
        "n",
        "Delta",
        "outliers",
        "seed",
    ]
    metrics_columns = [
        "count_reject",
        "count_fail_to_reject",
        "count_captured",
        "avg_width",
    ]

    # Check if cached simulation data exists
    filename = "dmeans_perf_metrics" \
                + "__ns_" + "_".join(map(str,ns)) \
                + "__Deltas_" + "_".join(map(str,Deltas)) \
                + "__outs_" + "_".join(map(str,outliers_options)) \
                + "__reps_" + str(reps) + ".csv"
    filepath = os.path.join("simdata", filename)
    if os.path.exists(filepath):
        print("loaded cached results from ", filepath)
        results = pd.read_csv(filepath, index_col=[0,1])
        return results
    
    # Prepare results table
    results_columns = dataset_columns + metrics_columns
    specs_idx = range(len(dataset_specs))
    models_idx = ["perm", "welch", "norm_bayes", "robust_bayes", "bf"]
    results_index = pd.MultiIndex.from_product((specs_idx, models_idx), names=["spec", "model"])
    results = pd.DataFrame(index=results_index, columns=results_columns)

    
    for row_idx, spec in enumerate(dataset_specs):
        print(spec)

        # Get rows that correspond to this `spec` (for all models)
        spec_results = results.loc[(row_idx,),:]

        # Save dataset spec columns
        spec_results.loc[:,"n"] = spec["n"]
        spec_results.loc[:,"Delta"] = spec["Delta"]
        spec_results.loc[:,"outliers"] = spec["outliers"]
        spec_results.loc[:,"seed"] = spec["random_seed"]
        # Initialize metrics columns
        spec_results.loc[:,"count_reject"] = 0
        spec_results.loc[:,"count_fail_to_reject"] = 0
        spec_results.loc[:,"count_captured"] = 0
        spec_results.loc["bf","count_captured"] = pd.NA
        spec_results.loc["bf","avg_width"] = pd.NA

        # Interpret the sparsity specification
        spec["outliers"] in ["no", "few", "lots"]
        if spec["outliers"] == "no":
            prop_outliers = 0
        elif spec["outliers"] == "few":
            prop_outliers = 0.02
        elif spec["outliers"] == "lots":
            prop_outliers = 0.05


        # Repeat for `rep` datasets
        ci_widths = defaultdict(list)
        t_before_reps = time.time()
        for rep in range(reps):
            if rep % 10 == 0:
                print("Running", spec, "rep", rep) 
            rep_random_seed = spec["random_seed"] * 10000 + rep
            dataset = gen_dmeans_dataset(n=spec["n"],
                                         Delta=spec["Delta"],
                                         prop_outliers=prop_outliers,
                                         random_seed=rep_random_seed)
            # Run all the tests
            rep_results = fit_dmeans_models(dataset, random_seed=rep_random_seed)
            for model, model_results in rep_results.items():
                # Hypothesis test metrics
                decision = model_results["decision"]
                if decision == "reject H0":
                    spec_results.loc[model, "count_reject"] += 1
                else:
                    spec_results.loc[model, "count_fail_to_reject"] += 1
                if model == "bf":
                    continue
                # Confidence interval metrics
                ci90 = model_results["ci90"]
                if is_inside(spec["Delta"], ci90):
                    spec_results.loc[model, "count_captured"] += 1
                ci90_width = ci90[1] - ci90[0]
                ci_widths[model].append(ci90_width)

        t_after_reps = time.time()
        reps_time = t_after_reps - t_before_reps
        print(reps, "reps took", reps_time, "to run...")

        for model, widths in ci_widths.items():
            spec_results.loc[model,"avg_width"] = np.mean(widths)

    # Cache results to avoid need to recompute
    results.to_csv(filepath)
    print("Saved file to " + filepath)
    return results




# Tables with success metrics
################################################################################

def get_perf_table_typeI(results):
    """
    Analysis of false positive results for all models.
    For all the datasets generated with Delta=0 (the null hypothesis is true)
    but it is possible the hypothesis test procedure will reject H0,
    which is a Type I error (false positive).
    """
    results = results.copy()

    # Infer the number of `reps` that were simulated for each spec
    first_result = results.iloc[0,:]
    reps = first_result["count_reject"] + first_result["count_fail_to_reject"]

    # Calculate the false positive rate for each (spec,model) combination
    results.loc[:,"false_positives"] = results.loc[:,"count_reject"] / reps

    # Select only relevant rows and columns
    subset_rows = results["Delta"] == 0  # when H0 is true
    subset_cols = ["n", "outliers", "false_positives"]
    subset = results.loc[subset_rows, subset_cols]
    
    # Reshape the data to prepare the Type I errors table
    tableA = subset.reset_index(level="spec", drop=True) \
                   .assign(outliers=pd.Categorical(subset["outliers"],
                                                   categories=["no", "few", "lots"],
                                                   ordered=True)) \
                   .pivot_table(index=["outliers", "n"],
                                columns="model",
                                values="false_positives",
                                observed=True) \
                   .reindex(columns=["perm", "welch", "norm_bayes", "robust_bayes", "bf"]) \
                   .sort_index()
    return tableA


def get_perf_table_power(results, show_all=True):
    """
    Analysis of power (true positives) for all models.
    For datasets Delta ≠ 0, correction decision is to reject H0.
    The `power` is the ability to the test to make the correct decision.
    """
    results = results.copy()

    # Infer the number of `reps` that were simulated for each spec
    first_result = results.iloc[0,:]
    reps = first_result["count_reject"] + first_result["count_fail_to_reject"]

    # Calculate the power for each (spec,model) combination
    results.loc[:,"power"] = 1 - results.loc[:,"count_fail_to_reject"] / reps

    # Select only relevant rows and columns
    cond1 = results["Delta"] != 0    # not part of the experiment
    cond2 = results["Delta"] != 0.2  # near 0% power
    cond3 = results["Delta"] != 1.3  # near 100% power
    cond4 = ~((results["Delta"] == 0.8) & (results["n"] == 100))  # near 100%
    cond5 = ~((results["Delta"] == 0.5) & (results["n"] == 20))   # near 0%
    if show_all:  # For display in notebooks
        subset_rows = cond1
    else:         # For reduced display in print book
        subset_rows = cond1 & cond2 & cond3 & cond4 & cond5
    subset_cols = ["n", "outliers", "Delta", "power"]
    subset = results.loc[subset_rows, subset_cols]
    
    # Reshape the data to prepare the power table
    tableB = subset.reset_index(level="spec", drop=True) \
                   .assign(outliers=pd.Categorical(subset["outliers"],
                                                   categories=["no", "few", "lots"],
                                                   ordered=True)) \
                   .pivot_table(index=["outliers", "Delta", "n"],
                                columns="model",
                                values="power",
                                observed=True) \
                   .reindex(columns=["perm", "welch", "norm_bayes", "robust_bayes", "bf"]) \
                   .sort_index()
    return tableB


def get_perf_table_coverage(results):
    """
    Calculates the summary of the coverage probability
    and width of the interval estimates from different models.
    """
    results = results.copy()

    # Infer the number of `reps` that were simulated for each spec
    first_result = results.iloc[0,:]
    reps = first_result["count_reject"] + first_result["count_fail_to_reject"]

    # Calculate the coverage proportion for each (spec,model) combination
    results.loc[:,"coverage"] = results.loc[:,"count_captured"] / reps

    # Select only relevant rows and columns
    subset_rows = results["Delta"] != 0
    subset_cols = ["n", "outliers", "Delta", "coverage", "avg_width"]
    subset = results.loc[subset_rows, subset_cols]
    # Drop the `bf` model rows because they don't have interval estimates
    bf_rows_to_drop = subset.loc[pd.IndexSlice[:,"bf"],:].index
    subset = subset.drop(index=bf_rows_to_drop)

    # Reshape the data to prepare the interval estimates table
    tableC = subset.reset_index(level="spec", drop=True) \
                   .assign(outliers=pd.Categorical(subset["outliers"],
                                                   categories=["no", "few", "lots"],
                                                   ordered=True)) \
                   .pivot_table(index=["outliers", "n", "Delta"],
                                columns="model",
                                values=["coverage", "avg_width"],
                                observed=True) \
                   .groupby(level=["outliers", "n"],
                            observed=True).mean() 
    tableC.columns = tableC.columns.swaplevel(0, 1)

    # Set the desired sort order of the columns index
    models = ["perm", "welch", "norm_bayes", "robust_bayes"]
    metrics = ["coverage", "avg_width"]
    new_cols = [(model, metric) for model in models for metric in metrics]
    tableC = tableC.loc[:, new_cols]
    return tableC

